#!/usr/bin/env python

"""
Unit test recorder functions.

"""

from __future__ import absolute_import, print_function

# Allows use of lots of single character variable names.
# pylint: disable=invalid-name,missing-docstring,too-many-statements
# pylint: disable=too-many-lines,no-self-use
import os
import os.path
import logging
import logging.handlers
from io import open as _open

import unittest2 as unittest  # we use @skip introduced in py27
import six
from testfixtures import LogCapture, log_capture
# Enabled only to display a tree of loggers
# from logging_tree import printout as logging_tree_printout
import yaml
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from pywbem import CIMInstanceName, CIMInstance, \
    Uint8, Uint16, Uint32, Uint64, Sint8, Sint16, \
    Sint32, Sint64, Real32, Real64, CIMProperty, CIMDateTime, CIMError, \
    HTTPError, WBEMConnection, LogOperationRecorder, configure_logger
# Renamed the following import to not have py.test pick it up as a test class:
from pywbem import TestClientRecorder as _TestClientRecorder
from pywbem.cim_operations import pull_path_result_tuple, pull_inst_result_tuple
from pywbem._utils import _format
from pywbem_mock import FakedWBEMConnection

from ..utils.dmtf_mof_schema_def import install_test_dmtf_schema


TEST_DIR = os.path.dirname(__file__)

# test outpuf file for the recorder tests.  This is opened for each
# test to save yaml output and may be reloaded during the same test
# to confirm the yaml results.
TEST_YAML_FILE = os.path.join(TEST_DIR, 'test_recorder.yaml')

TEST_OUTPUT_LOG = os.path.join(TEST_DIR, 'test_recorder.log')

VERBOSE = False


class BaseRecorderTests(unittest.TestCase):
    """Base class for recorder unit tests. Implements methods
       for creating instance and instancename
    """

    def create_ciminstance(self, set_path=False):
        """
        Create a sample instance with multiple properties and
        property types.
        """

        props_input = OrderedDict([
            ('S1', b'Ham'),
            ('Bool', True),
            ('UI8', Uint8(42)),
            ('UI16', Uint16(4216)),
            ('UI32', Uint32(4232)),
            ('UI64', Uint64(4264)),
            ('SI8', Sint8(-42)),
            ('SI16', Sint16(-4216)),
            ('SI32', Sint32(-4232)),
            ('SI64', Sint64(-4264)),
            ('R32', Real32(42.0)),
            ('R64', Real64(42.64)),
            ('S2', u'H\u00E4m'),  # U+00E4 = lower case a umlaut
            # CIMDateTime.__repr__() output changes between Python versions,
            # so we omit that from these properties. After all, this is not a
            # test for a correct repr() of all CIM datatypes, but a test of
            # the recorder with a long string.
        ])

        inst = CIMInstance('CIM_Foo', props_input)
        if set_path:
            inst.path = self.create_ciminstancename()
        return inst

    def create_ciminstancename(self):
        kb = [('Chicken', 'Ham')]
        obj_name = CIMInstanceName('CIM_Foo',
                                   kb,
                                   namespace='root/cimv2',
                                   host='woot.com')
        return obj_name

    def create_ciminstances(self):
        """Create multiple instances using """
        instances = []
        instances.append(self.create_ciminstance(set_path=True))
        instances.append(self.create_ciminstance(set_path=True))
        return instances

    def create_ciminstancepaths(self):
        """Create multiple instances using """
        paths = []
        paths.append(self.create_ciminstancename())
        paths.append(self.create_ciminstancename())
        return paths

    def create_method_params(self):
        """Create a set of method params to be used in InvokeMethodTests."""
        obj_name = self.create_ciminstancename()

        params = [('StringParam', 'Spotty'),
                  ('Uint8', Uint8(1)),
                  ('Sint8', Sint8(2)),
                  ('Uint16', Uint16(3)),
                  ('Sint16', Sint16(3)),
                  ('Uint32', Uint32(4)),
                  ('Sint32', Sint32(5)),
                  ('Uint64', Uint64(6)),
                  ('Sint64', Sint64(7)),
                  ('Real32', Real32(8)),
                  ('Real64', Real64(9)),
                  ('Bool', True),
                  ('Ref', obj_name)]
        # CIMDateTime.__repr__() output changes between Python
        # versions, so we omit any datetime parameters. After all,
        # this is not a test for a correct repr() of all CIM
        # datatypes, but a test of the recorder with a long string.
        return params


class ClientRecorderTests(BaseRecorderTests):
    """
    Common base for all tests on the TestClientRecorder. Defines specific common
    methods including setUp and tearDown for the TestClientRecorder.
    """
    def setUp(self):
        """ Setup recorder instance including defining output file"""
        self.testyamlfile = os.path.join(TEST_DIR, TEST_YAML_FILE)
        if os.path.isfile(self.testyamlfile):
            os.remove(self.testyamlfile)

        self.yamlfp = _TestClientRecorder.open_file(self.testyamlfile, 'a')

        self.test_recorder = _TestClientRecorder(self.yamlfp)
        self.test_recorder.reset()
        self.test_recorder.enable()

    def tearDown(self):
        """Close the test_client YAML file."""
        if self.yamlfp is not None:
            self.yamlfp.close()

    def closeYamlFile(self):
        """Close the yaml file if it is open"""
        if self.yamlfp is not None:
            self.yamlfp.close()
            self.yamlfp = None

    def loadYamlFile(self):
        """Load any created yaml file"""
        self.closeYamlFile()
        with _open(self.testyamlfile, encoding="utf-8") as fp:
            testyaml = yaml.safe_load(fp)
        return testyaml


class ToYaml(ClientRecorderTests):
    """Test the toyaml function with multiple data input"""
    def test_inst_to_yaml_simple(self):
        """Test Simple instancename toyaml conversion"""

        test_yaml = self.test_recorder.toyaml(self.create_ciminstancename())

        self.assertEqual(test_yaml['pywbem_object'], 'CIMInstanceName')
        self.assertEqual(test_yaml['classname'], 'CIM_Foo')
        self.assertEqual(test_yaml['namespace'], 'root/cimv2')
        kb = test_yaml['keybindings']
        self.assertEqual(kb['Chicken'], 'Ham')

        # CIMClass, cimqualifierdecl

    def test_to_yaml_simple2(self):
        """Test Simple cimdatetime and other primitive types to toyaml"""

        test_yaml = self.test_recorder.toyaml(
            CIMDateTime('20140924193040.654321+120'))

        self.assertEqual(test_yaml, '20140924193040.654321+120')

        self.assertEqual(self.test_recorder.toyaml(True), True)
        self.assertEqual(self.test_recorder.toyaml(False), False)
        self.assertEqual(self.test_recorder.toyaml(1234), 1234)
        self.assertEqual(self.test_recorder.toyaml('blahblah '), 'blahblah ')

    def test_inst_to_yaml_all_prop_types(self):
        """Test all property types toyaml"""

        inst = self.create_ciminstance()

        test_yaml = self.test_recorder.toyaml(inst)

        self.assertEqual(test_yaml['pywbem_object'], 'CIMInstance')
        self.assertEqual(test_yaml['classname'], 'CIM_Foo')
        properties = test_yaml['properties']

        si64 = properties['SI64']
        self.assertEqual(si64['name'], 'SI64')
        self.assertEqual(si64['type'], 'sint64')
        self.assertEqual(si64['value'], -4264)

        # TODO add test for reference properties
        # TODO add test for embedded object property

    def test_inst_to_yaml_array_props(self):
        """Test  property with array toyaml"""
        str_data = "The pink fox jumped over the big blue dog"
        array_props = [
            ('MyString', str_data),
            ('MyUint8Array', [Uint8(1), Uint8(2)]),
            ('MySint8Array', [Sint8(1), Sint8(2)]),
            ('MyUint64Array', [Uint64(123456789),
                               Uint64(123456789),
                               Uint64(123456789)]),
            ('MyUint32Array', [Uint32(9999), Uint32(9999)]),
            ('MyStrLongArray', [str_data, str_data, str_data]),
        ]
        inst = CIMInstance('CIM_FooArray', array_props)
        test_yaml = self.test_recorder.toyaml(inst)

        self.assertEqual(test_yaml['pywbem_object'], 'CIMInstance')
        self.assertEqual(test_yaml['classname'], 'CIM_FooArray')
        properties = test_yaml['properties']
        my_string = properties['MyString']
        self.assertEqual(my_string['name'], 'MyString')
        self.assertEqual(my_string['type'], 'string')
        self.assertEqual(my_string['value'], str_data)

        my_uint8array = properties['MyUint8Array']
        self.assertEqual(my_uint8array['name'], 'MyUint8Array')
        self.assertEqual(my_uint8array['type'], 'uint8')
        self.assertEqual(my_uint8array['value'], [Uint8(1), Uint8(2)])

        my_sint8array = properties['MySint8Array']
        self.assertEqual(my_sint8array['name'], 'MySint8Array')
        self.assertEqual(my_sint8array['type'], 'sint8')
        self.assertEqual(my_sint8array['value'], [Sint8(1), Sint8(2)])

        my_sint64array = properties['MyUint64Array']
        self.assertEqual(my_sint64array['name'], 'MyUint64Array')
        self.assertEqual(my_sint64array['type'], 'uint64')
        self.assertEqual(my_sint64array['value'], [Uint64(123456789),
                                                   Uint64(123456789),
                                                   Uint64(123456789)])

    def test_instname_to_yaml(self):
        """Test  instname toyaml"""

        inst_name = self.create_ciminstancename()

        test_yaml = self.test_recorder.toyaml(inst_name)

        self.assertEqual(test_yaml['pywbem_object'], 'CIMInstanceName')
        self.assertEqual(test_yaml['classname'], 'CIM_Foo')
        # TODO host does not appear in output yaml. Is that correct???
        # ##self.assertEqual[test_yaml['host'], 'woot.com']
        kbs = test_yaml['keybindings']
        self.assertEqual(len(kbs), 1)
        self.assertEqual(kbs['Chicken'], 'Ham')

    def test_openreq_resulttuple(self):
        """test tuple result from open operation. The input is a
        named tuple.
        """
        result = []
        context = ('test_rtn_context', 'root/cim_namespace')
        result_tuple = pull_path_result_tuple(result, True, None)

        test_yaml = self.test_recorder.toyaml(result_tuple)

        self.assertEqual(test_yaml['paths'], [])
        self.assertEqual(test_yaml['eos'], True)
        self.assertEqual(test_yaml['context'], None)

        result_tuple = pull_path_result_tuple(result, False, context)

        test_yaml = self.test_recorder.toyaml(result_tuple)

        self.assertEqual(test_yaml['paths'], [])
        self.assertEqual(test_yaml['eos'], False)
        self.assertEqual(test_yaml['context'], list(context))


class ClientOperationStageTests(ClientRecorderTests):
    """
    Test staging for different cim_operations.  This defines fixed
    parameters for the before and after staging, stages (which creates
    a yaml file), and then inspects that file to determine if valid
    yaml was created
    """

    def test_invoke_method(self):
        """
        Emulates call to invokemethod to test parameter processing.
        Currently creates the pywbem_request component.
        Each test emulated a single cim operation with fixed data to
        create the input for the yaml, create the yaml, and test the result
        """
        params = self.create_method_params()
        in_params_dict = OrderedDict(params)
        obj_name = in_params_dict['Ref']

        self.test_recorder.stage_pywbem_args(method='InvokeMethod',
                                             MethodName='Blah',
                                             ObjectName=obj_name,
                                             Params=params)
        method_result_tuple = None
        method_exception = None

        self.test_recorder.stage_pywbem_result(method_result_tuple,
                                               method_exception)

        self.test_recorder.record_staged()

        # reload the yaml to test created values
        test_yaml = self.loadYamlFile()
        test_yaml = test_yaml[0]

        pywbem_request = test_yaml['pywbem_request']
        self.assertEqual(pywbem_request['url'], 'http://acme.com:80')

        operation = pywbem_request['operation']
        self.assertEqual(operation['pywbem_method'], 'InvokeMethod')
        self.assertEqual(operation['MethodName'], 'Blah')
        out_params_dict = OrderedDict(operation['Params'])

        self.assertEqual(len(in_params_dict), len(out_params_dict))

        self.assertEqual(out_params_dict['StringParam'], 'Spotty')
        self.assertEqual(out_params_dict['Uint8'], 1)
        self.assertEqual(out_params_dict['Sint8'], 2)
        self.assertEqual(out_params_dict['Uint16'], 3)
        self.assertEqual(out_params_dict['Sint16'], 3)
        self.assertEqual(out_params_dict['Uint32'], 4)
        self.assertEqual(out_params_dict['Sint32'], 5)
        self.assertEqual(out_params_dict['Uint64'], 6)
        self.assertEqual(out_params_dict['Sint64'], 7)
        self.assertEqual(out_params_dict['Real32'], 8)
        self.assertEqual(out_params_dict['Real64'], 9)
        self.assertEqual(out_params_dict['Bool'], True)
        # TODO fix the following. Currently it fails with:
        # AssertionError: CIMIn[16 chars]name={'classname': 'CIM_Foo',
        # 'keybindings': {[132 chars]None) != CIMIn[16 chars]name=u'CIM_Foo',
        # keybindings=NocaseDict({('Chi[55 chars]com')
        # self.assertEqual(CIMInstanceName(out_params_dict['Ref']), obj_name)

    def test_get_instance(self):
        """
        Emulates call to getInstance to test parameter processing.
        Currently creates the pywbem_request component.
        """
        InstanceName = self.create_ciminstancename()

        self.test_recorder.reset()

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=InstanceName,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        instance = None
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        self.test_recorder.record_staged()

        test_yaml = self.loadYamlFile()
        test_yaml = test_yaml[0]
        pywbem_request = test_yaml['pywbem_request']
        self.assertEqual(pywbem_request['url'], 'http://acme.com:80')
        operation = pywbem_request['operation']
        self.assertEqual(operation['pywbem_method'], 'GetInstance')
        self.assertEqual(operation['LocalOnly'], True)
        self.assertEqual(operation['IncludeQualifiers'], True)
        self.assertEqual(operation['PropertyList'], ['propertyblah'])
        pywbem_response = test_yaml['pywbem_response']
        self.assertEqual(pywbem_response, {})

    def test_create_instance(self):
        """Test record of create instance"""

        NewInstance = self.create_ciminstance()

        self.test_recorder.reset()

        self.test_recorder.stage_pywbem_args(
            method='CreateInstance',
            NewInstance=NewInstance,
            namespace='cim/blah')

        exc = None
        obj_name = self.create_ciminstancename()

        self.test_recorder.stage_pywbem_result(obj_name, exc)

        self.test_recorder.record_staged()

        test_yaml = self.loadYamlFile()
        test_yaml = test_yaml[0]
        pywbem_request = test_yaml['pywbem_request']
        self.assertEqual(pywbem_request['url'], 'http://acme.com:80')
        operation = pywbem_request['operation']
        self.assertEqual(operation['pywbem_method'], 'CreateInstance')
        returned_new_inst = operation['NewInstance']
        self.assertEqual(returned_new_inst['classname'], NewInstance.classname)
        pd = returned_new_inst['properties']

        # compare all properties returned against original
        self.assertEqual(len(pd), len(NewInstance.properties))
        for pn, pv in pd.items():
            prop = CIMProperty(pn, pv['value'], type=pv['type'],
                               array_size=pv['array_size'],
                               propagated=pv['propagated'],
                               is_array=pv['is_array'],
                               reference_class=pv['reference_class'],
                               qualifiers=pv['qualifiers'],
                               embedded_object=pv['embedded_object'])

            self.assertEqual(NewInstance.properties[pn], prop,
                             'Property compare failed orig %s, recreated %s' %
                             (NewInstance.properties[pn], prop))

    def test_open_enumerateInstancePaths(self):
        """Emulate staging of enumerateInstancePaths call. Tests building
        yaml for this call. Used to test the generation of context, eos
        result info.
        """
        namespace = 'root/cimv2'

        self.test_recorder.reset(pull_op=True)

        self.test_recorder.stage_pywbem_args(
            method='OpenEnumerateInstancePaths',
            ClassName='CIM_BLAH',
            namespace=namespace,
            FilterQueryLanguage='WQL',
            FilterQuery='Property = 3',
            OperationTimeout=40,
            ContinueOnError=False,
            MaxObjectCount=100)

        exc = None
        result = []
        result_tuple = pull_path_result_tuple(result, True, None)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        self.test_recorder.record_staged()

        test_yaml = self.loadYamlFile()
        test_yaml = test_yaml[0]

        pywbem_request = test_yaml['pywbem_request']
        self.assertEqual(pywbem_request['url'], 'http://acme.com:80')
        operation = pywbem_request['operation']
        self.assertEqual(operation['pywbem_method'],
                         'OpenEnumerateInstancePaths')
        self.assertEqual(operation['ClassName'], 'CIM_BLAH')
        self.assertEqual(operation['MaxObjectCount'], 100)
        self.assertEqual(operation['FilterQueryLanguage'], 'WQL')
        self.assertEqual(operation['FilterQuery'], 'Property = 3')
        self.assertEqual(operation['ContinueOnError'], False)
        self.assertEqual(operation['OperationTimeout'], 40)

        pywbem_response = test_yaml['pywbem_response']
        pull_result = pywbem_response['pullresult']
        self.assertEqual(pull_result['paths'], [])
        self.assertEqual(pull_result['eos'], True)
        self.assertEqual(pull_result['context'], None)


################################################################
#
#           LogOperationRecorder tests
#
################################################################

class BaseLogOperationRecorderTests(BaseRecorderTests):
    """
    Test the LogOperationRecorder functions. Creates log entries and
    uses testfixture to validate results
    """
    def setUp(self):
        """
        Setup that is run before each test method.
        Shut down any existing logger and reset WBEMConnection and
        reset WBEMConnection class attributes
        """
        # pylint: disable=protected-access
        WBEMConnection._reset_logging_config()
        logging.shutdown()
        # NOTE We do not clean up handlers or logger names already defined.
        #      That should not affect the tests.

    def recorder_setup(self, detail_level=None):
        """Setup the recorder for a defined max output size"""

        configure_logger('api', log_dest='file',
                         detail_level=detail_level,
                         log_filename=TEST_OUTPUT_LOG, propagate=True)

        configure_logger('http', log_dest='file',
                         detail_level=detail_level,
                         log_filename=TEST_OUTPUT_LOG, propagate=True)

        # Define an attribute that is a single LogOperationRecorder to be used
        # in some of the tests.  Note that if detail_level is dict it is used
        # directly.
        if isinstance(detail_level, (six.string_types, six.integer_types)):
            detail_level = {'api': detail_level, 'http': detail_level}

        # pylint: disable=attribute-defined-outside-init
        # Set a conn id into the connection. Saves testing the connection
        # log for each test.
        self.test_recorder = LogOperationRecorder('test_id',
                                                  detail_levels=detail_level)
        # pylint: disable=protected-access
        self.test_recorder.reset()
        self.test_recorder.enable()

    def tearDown(self):
        """Remove LogCapture."""
        LogCapture.uninstall_all()
        logging.shutdown()
        # remove any existing log file
        if os.path.isfile(TEST_OUTPUT_LOG):
            os.remove(TEST_OUTPUT_LOG)


class LogOperationRecorderStagingTests(BaseLogOperationRecorderTests):
    """
    Test staging for different cim_operations.  This defines fixed
    parameters for the before and after staging, stages (which creates
    a yaml file), and then inspects that file to determine if valid
    yaml was created
    """

    @log_capture()
    def test_create_connection1(self, lc):
        """Create connection with default parameters"""
        # Fake the connection to create a fixed data environment
        conn = WBEMConnection('http://blah')
        # TODO AM 2018-06: Suppress the printing to stderr
        configure_logger('api', log_dest='file', detail_level='all',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace='root/cimv2', "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )

    @log_capture()
    def test_create_connection2(self, lc):
        """Test log of wbem connection with detailed information"""

        x509_dict = {"cert_file": 'Certfile.x', 'key_file': 'keyfile.x'}
        conn = WBEMConnection('http://blah',
                              default_namespace='root/blah',
                              creds=('username', 'password'),
                              x509=x509_dict,
                              no_verification=True,
                              timeout=10,
                              use_pull_operations=True,
                              stats_enabled=True)
        # TODO AM 2018-06: Suppress the printing to stderr
        configure_logger('api', log_dest='file', detail_level='all',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah', "
            "creds=('username', ...), "
            "conn_id={0!A}, "
            "default_namespace='root/blah', "
            "x509={{'cert_file': 'Certfile.x', 'key_file': 'keyfile.x'}}, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=True, "
            "timeout=10, "
            "use_pull_operations=True, "
            "stats_enabled=True, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )

    @log_capture()
    def test_create_connection_summary(self, lc):
        """Test log of wbem connection with detailed information"""

        x509_dict = {"cert_file": 'Certfile.x', 'key_file': 'keyfile.x'}
        conn = WBEMConnection('http://blah',
                              default_namespace='root/blah',
                              creds=('username', 'password'),
                              x509=x509_dict,
                              no_verification=True,
                              timeout=10,
                              use_pull_operations=True,
                              stats_enabled=True)
        # TODO AM 2018-06: Suppress the printing to stderr
        configure_logger('api', log_dest='file', detail_level='summary',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah', "
            "creds=('username', ...), "
            "default_namespace='root/blah', "
            "...)",
            conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )

    @log_capture()
    def test_stage_result_exception(self, lc):
        """Test the ops result log None return, HTTPError exception."""
        self.recorder_setup(detail_level=10)

        # Note: cimerror is the CIMError HTTP header field
        exc = HTTPError(500, "Fake Reason", cimerror="Fake CIMError")

        self.test_recorder.stage_pywbem_result(None, exc)

        result_exc = "Exception:test_id None('HTTPError...)"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    @log_capture()
    def test_stage_result_exception_all(self, lc):
        """Test the ops result log None return, HTTPError exception."""
        self.recorder_setup(detail_level='all')

        # Note: cimerror is the CIMError HTTP header field
        exc = HTTPError(500, "Fake Reason", cimerror="Fake CIMError")

        self.test_recorder.stage_pywbem_result(None, exc)

        result_exc = _format(
            "Exception:test_id None('HTTPError({0})')",
            exc)

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    @log_capture()
    def test_stage_getinstance_args(self, lc):
        """
        Emulates call to getInstance to test parameter processing.
        Currently creates the pywbem_request component.
        """

        inst_name = self.create_ciminstancename()

        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=inst_name,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        result_req = (
            "Request:test_id GetInstance("
            "IncludeClassOrigin=True, "
            "IncludeQualifiers=True, "
            "InstanceName=CIMInstanceName("
            "classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', host='woot.com'), "
            "LocalOnly=True, "
            "PropertyList=['propertyblah'])"
        )

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
        )

    @log_capture()
    def test_stage_instance_result(self, lc):
        instance = self.create_ciminstance()
        self.recorder_setup(detail_level=10)
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_ret = "Return:test_id None(CIMInstanc...)"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_stage_instance_result_default(self, lc):
        instance = self.create_ciminstance()
        # set the length in accord with the "min" definition.
        self.recorder_setup(detail_level=1000)
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_ret = (
            "Return:test_id None(CIMInstance(classname='CIM_Foo', "
            "path=None, properties=NocaseDict({"
            "'S1': CIMProperty(name='S1', value='Ham', type='string', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'Bool': CIMProperty(name='Bool', value=True, "
            "type='boolean', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'UI8': CIMProperty(name='UI8', value=42, type='uint8', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI16': CIMProperty(name='UI16', value=4216, "
            "type='uint16', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'UI32': CIMProperty(name='UI32', value=4232, "
            "type='uint32', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None,...)")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_stage_instance_result_all(self, lc):
        instance = self.create_ciminstance()
        self.recorder_setup(detail_level='all')
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_ret = (
            "Return:test_id None(CIMInstance(classname='CIM_Foo', "
            "path=None, properties=NocaseDict({"
            "'S1': CIMProperty(name='S1', value='Ham', "
            "type='string', reference_class=None, "
            "embedded_object=None, is_array=False, array_size=None, "
            "class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'Bool': CIMProperty(name='Bool', value=True, "
            "type='boolean', reference_class=None, "
            "embedded_object=None, is_array=False, array_size=None, "
            "class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI8': CIMProperty(name='UI8', value=42, "
            "type='uint8', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'UI16': CIMProperty(name='UI16', value=4216, "
            "type='uint16', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'UI32': CIMProperty(name='UI32', value=4232, "
            "type='uint32', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'UI64': CIMProperty(name='UI64', value=4264, "
            "type='uint64', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI8': CIMProperty(name='SI8', value=-42, "
            "type='sint8', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI16': CIMProperty(name='SI16', value=-4216, "
            "type='sint16', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI32': CIMProperty(name='SI32', value=-4232, type='sint32', "
            "reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI64': CIMProperty(name='SI64', value=-4264, type='sint64', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'R32': CIMProperty(name='R32', value=42.0, type='real32', "
            "reference_class=None, embedded_object=None, is_array=False,"
            " array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'R64': CIMProperty(name='R64', value=42.64, type='real64', "
            "reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'S2': CIMProperty(name='S2', value='H\\u00e4m', type='string', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({}))"
            "}), property_list=None, qualifiers=NocaseDict({})))")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_stage_instance_result_inst_path(self, lc):
        instance = self.create_ciminstance(set_path=True)
        self.recorder_setup(detail_level='paths')
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_ret = (
            "Return:test_id None("
            "'//woot.com/root/cimv2:CIM_Foo.Chicken=\"Ham\"')")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_stage_instance_result_paths(self, lc):
        self.recorder_setup(detail_level='paths')
        instances = self.create_ciminstances()
        exc = None

        self.test_recorder.stage_pywbem_result(instances, exc)

        result_ret = (
            "Return:test_id None("
            "'//woot.com/root/cimv2:CIM_Foo.Chicken=\"Ham\"', "
            "'//woot.com/root/cimv2:CIM_Foo.Chicken=\"Ham\"')")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_stage_instance_result_pull(self, lc):
        self.recorder_setup(detail_level='paths')
        instances = self.create_ciminstances()
        exc = None

        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_inst_result_tuple(instances, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        result_ret = (
            "Return:test_id None(pull_inst_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "instances='//woot.com/root/cimv2:CIM_Foo.Chicken=\"Ham\"', "
            "'//woot.com/root/cimv2:CIM_Foo.Chicken=\"Ham\"'))"
        )

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def build_http_request(self):
        """
        Build an http request for the following tests
        """
        headers = OrderedDict([
            ('CIMOperation', 'MethodCall'),
            ('CIMMethod', 'GetInstance'),
            ('CIMObject', 'root/cimv2')])
        url = 'http://blah'
        method = 'POST'
        target = '/cimom'

        payload = (
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            '<SIMPLEREQ>\n'
            '<IMETHODCALL NAME="GetInstance">\n'
            '<LOCALNAMESPACEPATH>\n'
            '<NAMESPACE NAME="root"/>\n'
            '<NAMESPACE NAME="cimv2"/>\n'
            '</LOCALNAMESPACEPATH>\n'
            '<IPARAMVALUE NAME="InstanceName">\n'
            '<INSTANCENAME CLASSNAME="PyWBEM_Person">\n'
            '<KEYBINDING NAME="Name">\n'
            '<KEYVALUE VALUETYPE="string">Fritz</KEYVALUE>\n'
            '</KEYBINDING>\n'
            '</INSTANCENAME>\n'
            '</IPARAMVALUE>\n'
            '<IPARAMVALUE NAME="LocalOnly">\n'
            '<VALUE>FALSE</VALUE>\n'
            '</IPARAMVALUE>\n'
            '</IMETHODCALL>\n'
            '</SIMPLEREQ>\n'
            '</MESSAGE>\n'
            '</CIM>)')

        return url, target, method, headers, payload

    @log_capture()
    def test_stage_http_request_all(self, lc):
        """Test stage of http_request log with detail_level='all'"""
        self.recorder_setup(detail_level='all')
        url, target, method, headers, payload = self.build_http_request()

        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2'\n"
            '    <?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            '<SIMPLEREQ>\n'
            '<IMETHODCALL NAME="GetInstance">\n'
            '<LOCALNAMESPACEPATH>\n'
            '<NAMESPACE NAME="root"/>\n'
            '<NAMESPACE NAME="cimv2"/>\n'
            '</LOCALNAMESPACEPATH>\n'
            '<IPARAMVALUE NAME="InstanceName">\n'
            '<INSTANCENAME CLASSNAME="PyWBEM_Person">\n'
            '<KEYBINDING NAME="Name">\n'
            '<KEYVALUE VALUETYPE="string">Fritz</KEYVALUE>\n'
            '</KEYBINDING>\n'
            '</INSTANCENAME>\n'
            '</IPARAMVALUE>\n'
            '<IPARAMVALUE NAME="LocalOnly">\n'
            '<VALUE>FALSE</VALUE>\n'
            '</IPARAMVALUE>\n'
            '</IMETHODCALL>\n'
            '</SIMPLEREQ>\n'
            '</MESSAGE>\n'
            '</CIM>)')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )

    @log_capture()
    def test_stage_http_request_summary(self, lc):
        """
        Test http request log record with summary as detail level
        """
        self.recorder_setup(detail_level='summary')
        url, target, method, headers, payload = self.build_http_request()
        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2'\n"
            '    ')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )

    @log_capture()
    def test_stage_http_request_int(self, lc):
        """
        Test http log record with integer as detail_level
        """
        self.recorder_setup(detail_level=10)

        url, target, method, headers, payload = self.build_http_request()

        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2'\n"
            '    <?xml vers...')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )

    def build_http_response(self):
        """
        Build an http response. Builds the complete stage_http_response1
        and executes and returns the body component so each test can build
        stage_http_response2 since that is where the logging occurs
        """
        body = ('<?xml version="1.0" encoding="utf-8" ?>\n'
                '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
                '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
                '<SIMPLERSP>\n'
                '<IMETHODRESPONSE NAME="GetInstance">\n'
                '<IRETURNVALUE>\n'
                '<INSTANCE CLASSNAME="PyWBEM_Person">\n'
                '<PROPERTY NAME="Name" TYPE="string">\n'
                '<VALUE>Fritz</VALUE>\n'
                '</PROPERTY>\n'
                '<PROPERTY NAME="Address" TYPE="string">\n'
                '<VALUE>Fritz Town</VALUE>\n'
                '</PROPERTY>\n'
                '</INSTANCE>\n'
                '</IRETURNVALUE>\n'
                '</IMETHODRESPONSE>\n'
                '</SIMPLERSP>\n'
                '</MESSAGE>\n'
                '</CIM>)\n')
        headers = OrderedDict([
            ('Content-type', 'application/xml; charset="utf-8"'),
            ('Content-length', str(len(body)))])
        status = 200
        reason = ""
        version = 11
        self.test_recorder.stage_http_response1('test_id', version,
                                                status, reason, headers)

        return body

    @log_capture()
    def test_stage_http_response_all(self, lc):
        """
        Test http response log record with 'all' detail_level
        """
        self.recorder_setup(detail_level='all')

        body = self.build_http_response()

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'450'\n"
            '    <?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            '<SIMPLERSP>\n'
            '<IMETHODRESPONSE NAME="GetInstance">\n'
            '<IRETURNVALUE>\n'
            '<INSTANCE CLASSNAME="PyWBEM_Person">\n'
            '<PROPERTY NAME="Name" TYPE="string">\n'
            '<VALUE>Fritz</VALUE>\n'
            '</PROPERTY>\n'
            '<PROPERTY NAME="Address" TYPE="string">\n'
            '<VALUE>Fritz Town</VALUE>\n'
            '</PROPERTY>\n'
            '</INSTANCE>\n'
            '</IRETURNVALUE>\n'
            '</IMETHODRESPONSE>\n'
            '</SIMPLERSP>\n'
            '</MESSAGE>\n'
            '</CIM>)\n')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )

    @log_capture()
    def test_stage_http_response_summary(self, lc):
        """
        Test http response log record with 'all' detail_level
        """
        self.recorder_setup(detail_level='summary')

        body = self.build_http_response()

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'450'\n"
            '    ')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )

    @log_capture()
    def test_stage_http_response_int(self, lc):
        """
        Test http response log record with 'all' detail_level
        """
        self.recorder_setup(detail_level=30)

        body = self.build_http_response()

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'450'\n"
            '    <?xml version="1.0" encoding="...')

        lc.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )


class LogOperationRecorderTests(BaseLogOperationRecorderTests):
    """
    Test args and results logging. This emulates the WBEMConnection method
    call and the response together.
    """

    @log_capture()
    def test_getinstance(self, lc):
        """Test the ops result log for get instance"""

        inst_name = self.create_ciminstancename()

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=inst_name,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])
        instance = self.create_ciminstance()
        exc = None
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance(IncludeCla...)")

        result_ret = (
            "Return:test_id GetInstance(CIMInstanc...)")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_getinstance_exception(self, lc):
        """Test the ops result log for get instance"""

        inst_name = self.create_ciminstancename()

        self.recorder_setup(detail_level=11)

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=inst_name,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])
        instance = None
        exc = CIMError(6, "Fake CIMError")
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance(IncludeClas...)")

        result_exc = (
            "Exception:test_id GetInstance('CIMError(6...)")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    @log_capture()
    def test_getinstance_exception_all(self, lc):
        """Test the ops result log for get instance CIMError exception"""

        inst_name = self.create_ciminstancename()

        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=inst_name)
        instance = None
        exc = CIMError(6, "Fake CIMError")
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance("
            "InstanceName=CIMInstanceName("
            "classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', host='woot.com'))")

        result_exc = _format(
            "Exception:test_id GetInstance('CIMError({0})')", exc)

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    @log_capture()
    def test_getinstance_result_all(self, lc):
        """Test the ops result log for get instance"""

        inst_name = self.create_ciminstancename()

        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=inst_name,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])
        instance = self.create_ciminstance()
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            'Request:test_id GetInstance(IncludeClassOrigin=True, '
            'IncludeQualifiers=True, '
            "InstanceName=CIMInstanceName(classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', "
            "host='woot.com'), LocalOnly=True, "
            "PropertyList=['propertyblah'])")

        result_ret = (
            "Return:test_id GetInstance(CIMInstance(classname='CIM_Foo', "
            "path=None, properties=NocaseDict({"
            "'S1': CIMProperty(name='S1', value='Ham', type='string', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'Bool': CIMProperty(name='Bool', value=True, type='boolean', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI8': CIMProperty(name='UI8', value=42, type='uint8', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI16': CIMProperty(name='UI16', value=4216, type='uint16', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI32': CIMProperty(name='UI32', value=4232, type='uint32', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'UI64': CIMProperty(name='UI64', value=4264, "
            "type='uint64', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI8': CIMProperty(name='SI8', value=-42, type='sint8', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'SI16': CIMProperty(name='SI16', value=-4216, type='sint16', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'SI32': CIMProperty(name='SI32', value=-4232, "
            "type='sint32', reference_class=None, embedded_object=None, "
            "is_array=False, array_size=None, class_origin=None, "
            "propagated=None, qualifiers=NocaseDict({})), "
            "'SI64': CIMProperty(name='SI64', value=-4264, type='sint64', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'R32': CIMProperty(name='R32', value=42.0, type='real32', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'R64': CIMProperty(name='R64', value=42.64, type='real64', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({})), "
            "'S2': CIMProperty(name='S2', value='H\\u00e4m', type='string', "
            "reference_class=None, embedded_object=None, is_array=False, "
            "array_size=None, class_origin=None, propagated=None, "
            "qualifiers=NocaseDict({}))"
            "}), property_list=None, qualifiers=NocaseDict({})))")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_enuminstances_result(self, lc):
        """Test the ops result log for enumerate instances"""

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstances',
            ClassName='CIM_Foo',
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        instance = self.create_ciminstance()
        exc = None

        self.test_recorder.stage_pywbem_result([instance, instance], exc)

        result_req = (
            "Request:test_id EnumerateInstances(ClassName=...)")

        result_ret = (
            "Return:test_id EnumerateInstances([CIMInstan...)")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_enuminstancenames_result(self, lc):
        """Test the ops result log for enumerate instances"""

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstanceNames',
            ClassName='CIM_Foo',
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'blah2'])

        exc = None
        inst_name = self.create_ciminstancename()

        self.test_recorder.stage_pywbem_result([inst_name, inst_name], exc)

        result_req = (
            "Request:test_id EnumerateInstanceNames(ClassName=...)")

        result_ret = (
            "Return:test_id EnumerateInstanceNames([CIMInstan...)")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_openenuminstances_result_all(self, lc):
        """Test the ops result log for enumerate instances. Returns no
        instances.
        """

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='OpenEnumerateInstances',
            ClassName='CIM_Foo',
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        # instance = self.create_ciminstance()
        exc = None

        result = []
        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_inst_result_tuple(result, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        result_req = (
            "Request:test_id OpenEnumerateInstances("
            "ClassName='CIM_Foo', "
            "IncludeClassOrigin=True, "
            "IncludeQualifiers=True, "
            "LocalOnly=True, "
            "PropertyList=['propertyblah'])")

        result_ret = (
            "Return:test_id OpenEnumerateInstances(pull_inst_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "instances=[]))")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_openenuminstances_all(self, lc):
        """Test the ops result log for enumerate instances paths with
        data in the paths component"""

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='OpenEnumerateInstancePaths',
            ClassName='CIM_Foo',
            FilterQueryLanguage='FQL',
            FilterQuery='SELECT A from B',
            OperationTimeout=10,
            ContinueOnError=None,
            MaxObjectCount=100)

        inst_name = self.create_ciminstancename()
        result = [inst_name, inst_name]
        exc = None

        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_path_result_tuple(result, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        result_req = (
            "Request:test_id OpenEnumerateInstancePaths("
            "ClassName='CIM_Foo', "
            "ContinueOnError=None, "
            "FilterQuery='SELECT A from B', "
            "FilterQueryLanguage='FQL', "
            "MaxObjectCount=100, "
            "OperationTimeout=10)")

        result_ret = (
            "Return:test_id OpenEnumerateInstancePaths("
            "pull_path_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "paths=["
            "CIMInstanceName(classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', host='woot.com'), "
            "CIMInstanceName(classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', host='woot.com')"
            "]))")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_associators_result(self, lc):
        """Test the ops result log for Associators that returns nothing"""

        inst_name = self.create_ciminstancename()

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='Associators',
            InstanceName=inst_name,
            AssocClass='BLAH_Assoc',
            ResultClass='BLAH_Result',
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'propertyblah2'])
        exc = None

        self.test_recorder.stage_pywbem_result([], exc)

        result_req = (
            "Request:test_id Associators(AssocClass...)")

        result_ret = (
            "Return:test_id Associators([])")

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_associators_result_exception(self, lc):
        """Test the ops result log for associators that returns exception"""

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=11)

        exc = CIMError(6, "Fake CIMError")

        self.test_recorder.stage_pywbem_result([], exc)

        result_exc = "Exception:test_id None('CIMError(6...)"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    @log_capture()
    def test_invokemethod_int(self, lc):
        """Test invoke method log"""
        self.recorder_setup(detail_level=11)

        obj_name = self.create_ciminstancename()
        return_val = 0
        params = [('StringParam', 'Spotty'),
                  ('Uint8', Uint8(1)),
                  ('Sint8', Sint8(2))]

        self.test_recorder.stage_pywbem_args(method='InvokeMethod',
                                             MethodName='Blah',
                                             ObjectName=obj_name,
                                             Params=OrderedDict(params))

        self.test_recorder.stage_pywbem_result((return_val, params),
                                               None)

        result_req = "Request:test_id InvokeMethod(MethodName=...)"

        result_ret = "Return:test_id InvokeMethod((0, [('Stri...)"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @log_capture()
    def test_invokemethod_summary(self, lc):
        """Test invoke method log"""
        self.recorder_setup(detail_level='summary')

        obj_name = self.create_ciminstancename()
        return_val = 0
        params = [('StringParam', 'Spotty'),
                  ('Uint8', Uint8(1)),
                  ('Sint8', Sint8(2))]

        self.test_recorder.stage_pywbem_args(method='InvokeMethod',
                                             MethodName='Blah',
                                             ObjectName=obj_name,
                                             Params=OrderedDict(params))

        self.test_recorder.stage_pywbem_result((return_val, params),
                                               None)

        result_req = (
            "Request:test_id InvokeMethod("
            "MethodName='Blah', "
            "ObjectName=CIMInstanceName("
            "classname='CIM_Foo', "
            "keybindings=NocaseDict({'Chicken': 'Ham'}), "
            "namespace='root/cimv2', host='woot.com'), "
            "Params=OrderedDict(["
            "('StringParam', 'Spotty'), "
            "('Uint8', Uint8(cimtype='uint8', minvalue=0, maxvalue=255, 1)), "
            "('Sint8', Sint8(cimtype='sint8', minvalue=-128, maxvalue=127, 2))"
            "]))")

        result_ret = "Return:test_id InvokeMethod(tuple )"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    # TODO add tests for all for invoke method.


class TestExternLoggerDef(BaseLogOperationRecorderTests):
    """ Test configuring loggers above level of our loggers"""

    @log_capture()
    def test_root_logger(self, lc):
        """
        Create a logger using logging.basicConfig and generate logs
        """
        logging.basicConfig(filename=TEST_OUTPUT_LOG, level=logging.DEBUG)

        detail_level_int = 10
        detail_level = {'api': detail_level_int, 'http': detail_level_int}

        # pylint: disable=attribute-defined-outside-init
        self.test_recorder = LogOperationRecorder('test_id',
                                                  detail_levels=detail_level)

        # We don't activate because we are not using wbemconnection, just
        # the recorder calls.
        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstanceNames',
            ClassName='CIM_Foo',
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'blah2'])

        exc = None
        inst_name = self.create_ciminstancename()

        self.test_recorder.stage_pywbem_result([inst_name, inst_name], exc)

        result_req = "Request:test_id EnumerateInstanceNames(ClassName=...)"

        result_ret = "Return:test_id EnumerateInstanceNames([CIMInstan...)"

        lc.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @unittest.skip("Test unreliable exception not always the same")
    @log_capture()
    def test_pywbem_logger(self, lc):
        """
        Test executing a connection with an externally created handler. Note
        that this test inconsistently reports the text of the exception
        in that sometimes adds the message  "[Errno 113] No route to host" and
        sometimes the message "timed out".
        """
        logger = logging.getLogger('pywbem')
        handler = logging.FileHandler(TEST_OUTPUT_LOG)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        configure_logger('api', detail_level='summary', connection=True,
                         propagate=True)
        try:
            conn = WBEMConnection('http://blah', timeout=1)

            conn.GetClass('blah')
        except Exception:  # pylint: disable=broad-except
            pass

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)
        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_exc = _format(
            "Exception:{0} GetClass('ConnectionError(Socket error: "
            "[Errno 113] No route to host)')",
            conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace='root/cimv2', "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=1, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        result_req = _format(
            "Request:{0} GetClass("
            "ClassName='blah', "
            "IncludeClassOrigin=None, "
            "IncludeQualifiers=None, "
            "LocalOnly=None, "
            "PropertyList=None, "
            "namespace=None)",
            conn_id)

        result_hreq = _format(
            "Request:{0} POST /cimom 11 http://blah "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetClass' "
            "CIMObject:u'root/cimv2'\n"
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">'
            '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">'
            '<SIMPLEREQ>'
            '<IMETHODCALL NAME="GetClass">'
            '<LOCALNAMESPACEPATH>'
            '<NAMESPACE NAME="root"/>'
            '<NAMESPACE NAME="cimv2"/>'
            '</LOCALNAMESPACEPATH>'
            '<IPARAMVALUE NAME="ClassName">'
            '<CLASSNAME NAME="blah"/>'
            '</IPARAMVALUE>'
            '</IMETHODCALL>'
            '</SIMPLEREQ>'
            '</MESSAGE>'
            '</CIM>',
            conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (http_exp_log_id, 'DEBUG', result_hreq),
            (api_exp_log_id, 'DEBUG', result_exc),
        )


class TestLoggingEndToEnd(BaseLogOperationRecorderTests):
    """
    Test the examples documented in _logging.py document string. Because
    there is not logging until the first http request, these extend the
    examples to actually issue a request.  Therefore, they use the mocker
    to emulate a server

    """

    def build_repo(self, namespace):
        """Build a fake repo and FakeWBEMConnection so we get responses back"""
        schema = install_test_dmtf_schema()
        partial_schema = """
            #pragma locale ("en_US")
            #pragma include ("Interop/CIM_ObjectManager.mof")
            """

        conn = FakedWBEMConnection(default_namespace=namespace)
        conn.compile_mof_string(partial_schema, namespace=namespace,
                                search_paths=[schema.schema_mof_dir])
        return conn

    @log_capture()
    def test_1(self, lc):
        """
        Configure the "pywbem.api" logger for summary information output to a
        file and activate that logger for all subsequently created
        :class:`~pywbem.WBEMConnection` objects.
        NOTE: We changed from example to log to file and use log_captyre
        """
        # setup schema in test because we configure before we create the
        # connection in this test.
        namespace = 'root/blah'
        schema = install_test_dmtf_schema()
        partial_schema = """
            #pragma locale ("en_US")
            #pragma include ("Interop/CIM_ObjectManager.mof")
            """
        conn = FakedWBEMConnection(default_namespace=namespace)
        conn.compile_mof_string(partial_schema, namespace=namespace,
                                search_paths=[schema.schema_mof_dir])

        configure_logger('api', log_dest='file',
                         detail_level='summary',
                         log_filename=TEST_OUTPUT_LOG,
                         connection=conn, propagate=True)
        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "default_namespace={1!A}, ...))",
            conn_id, namespace)

        result_req = _format(
            "Request:{0} GetClass("
            "ClassName='CIM_ObjectManager', "
            "IncludeClassOrigin=None, "
            "IncludeQualifiers=None, "
            "LocalOnly=None, "
            "PropertyList=None, "
            "namespace={1!A})",
            conn_id, namespace)

        result_ret = _format(
            "Return:{0} GetClass(CIMClass CIM_ObjectManager)",
            conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret)
        )

    @log_capture()
    def test_2(self, lc):
        """
        Configure and activate a single :class:`~pywbem.WBEMConnection` object
        logger for output of summary information for both "pywbem.api" and
        "pywbem.http"::
        Differs from example in that we set detail_level to limit output for
        test
        """
        namespace = 'root/blah'
        conn = self.build_repo(namespace)
        configure_logger('all', log_dest='file',
                         log_filename=TEST_OUTPUT_LOG,
                         detail_level=10,
                         connection=conn, propagate=True)
        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    @log_capture()
    def test_3(self, lc):
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test.
        Does not produce http request/response info because when using the
        pywbem_mock, no http is generated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('all', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    @log_capture()
    def test_4(self, lc):
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test
        """
        logging.basicConfig(filename='basicconfig.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('api', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    @log_capture()
    def test_5(self, lc):
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test. The basic logger should have no effect since pywbem sets the
        NULLHandler at the pywbem level.
        """
        logging.basicConfig(filename='basicconfig.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('api', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        lc.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret)
        )

    @log_capture()
    def test_6(self, lc):
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest='file',
                         log_filename=TEST_OUTPUT_LOG,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        lc.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    @log_capture()
    def test_7(self, lc):
        """
        Configure a http logger with detail_level='all', log_dest=none
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        lc.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    @log_capture()
    def test_8(self, lc):
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        msg_format = '%(asctime)s-%(name)s-%(message)s'
        handler = logging.FileHandler(TEST_OUTPUT_LOG, encoding="UTF-8")
        handler.setFormatter(logging.Formatter(msg_format))
        logger = logging.getLogger('pywbem')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        lc.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    @log_capture()
    def test_9(self, lc):
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """

        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        lc.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    @log_capture()
    def test_10(self, lc):
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """

        namespace = 'root/blah'
        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all',
                         connection=True, propagate=True)
        conn = self.build_repo(namespace)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "verify_callback=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        lc.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    @log_capture()
    def test_err(self, lc):
        """
        Test configure_logger exception
        """
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        try:
            configure_logger('api', detail_level='blah', connection=conn,
                             propagate=True)
            self.fail("Exception expected")
        except ValueError:
            pass

        lc.check()


if __name__ == '__main__':
    unittest.main()



#
# (C) Copyright 2003-2005 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006-2007 Novell, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
# Author: Martin Pool <mbp@hp.com>
# Author: Bart Whiteley <bwhiteley@suse.de>
# Author: Ross Peoples <ross.peoples@gmail.com>
#

'''
Send HTTP/HTTPS requests to a WBEM server.

This module does not know anything about the fact that the data being
transferred in the HTTP request and response is CIM-XML.  It is up to the
caller to provide CIM-XML formatted input data and interpret the result data
as CIM-XML.
'''

from __future__ import print_function, absolute_import

import re
import os
import sys
import errno
import socket
import getpass
from stat import S_ISSOCK
import platform
import base64
import threading
from datetime import datetime
import warnings

import six
from six.moves import http_client as httplib
from six.moves import urllib

from .cim_obj import CIMClassName, CIMInstanceName
from .exceptions import ConnectionError, AuthError, TimeoutError, HTTPError
from ._warnings import ToleratedServerIssueWarning
from ._utils import _ensure_unicode, _ensure_bytes, _format

_ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'

if six.PY2 and not _ON_RTD:  # RTD has no swig to install M2Crypto
    # pylint: disable=wrong-import-order
    from M2Crypto import SSL           # pylint: disable=wrong-import-position
    from M2Crypto.Err import SSLError  # pylint: disable=wrong-import-position
    _HAVE_M2CRYPTO = True
    # pylint: disable=invalid-name
    SocketErrors = (socket.error, socket.sslerror)
else:
    import ssl as SSL                  # pylint: disable=wrong-import-position
    # pylint: disable=wrong-import-position
    from ssl import SSLError, CertificateError
    _HAVE_M2CRYPTO = False
    # pylint: disable=invalid-name
    SocketErrors = (socket.error,)

__all__ = ['DEFAULT_CA_CERT_PATHS']


def create_pywbem_ssl_context():
    """ Create an SSL context based on what is commonly accepted as the
        required limitations. This code attempts to create the same context for
        Python 2 and Python 3 except for the ciphers
        This list is based on what is currently defined in the Python SSL
        module create_default_context function
        This includes:

            * Disallow SSLV2 and SSLV3
            * Allow TLSV1 TLSV1.1, TLSV1.2
            * No compression
            * Single DH Use and Single ECDH use
        cacerts info is set independently so is not part of our context setter.
    """

    if six.PY2:
        context = SSL.Context('sslv23')
        # Many of the flags are not in the M2Crypto source so they were taken
        # from OpenSSL SSL.h module as flags.
        SSL.context.set_options(SSL.SSL_OP_NO_SSLv2 |
                                0x02000000 |  # OP_NO_SSLV3
                                0x00020000 |  # OP_NO_COMPRESSION
                                0x00100000 |  # OP_SINGLE_DH_USE
                                0x00400000 |  # OP_CIPHER_SERVER_PREFERENCE
                                0x00080000)   # OP_SINGLE_ECDH_USE
    else:
        # The choice for the Python SSL module is whether to use the
        # create_default directly and possibly have different limits depending
        # on which version of Python you use or to set the attributes
        # directly based on a currently used SSL
        context = SSL.create_default_context(purpose=SSL.Purpose.CLIENT_AUTH)

        # Variable settings per SSL create_default_context. These are what
        # the function above sets for Python 3.4
        # context = SSLContext(PROTOCOL_SSLv23)
        # context.options |= OP_NO_SSLv2
        # context.options |= OP_NO_SSLv3
        # context.options |= getattr(SSL, "OP_NO_COMPRESSION", 0)
        # context.options |= getattr(SSL, "OP_CIPHER_SERVER_PREFERENCE", 0)
        # context.options |= getattr(SSL, "OP_SINGLE_DH_USE", 0)
        # context.options |= getattr(SSL, "OP_SINGLE_ECDH_USE", 0)
        # context.set_ciphers(_RESTRICTED_SERVER_CIPHERS)

    return context


DEFAULT_PORT_HTTP = 5988        # default port for http
DEFAULT_PORT_HTTPS = 5989       # default port for https


#: Default directory paths to be used when the `ca_certs` parameter of
#: :class:`~pywbem.WBEMConnection` is `None`. The first existing directory is
#: used as a default for that parameter.
#: Note that these default directory paths only work on some Linux
#: distributions.
DEFAULT_CA_CERT_PATHS = \
    ['/etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt',
     '/etc/ssl/certs', '/etc/ssl/certificates']


def get_default_ca_cert_paths():
    """Return the list of default certificate paths defined for this
       system environment. This is the list of directories that
       should be searched to find a directory that contains
       certificates possibly suitable for the ssl ca_certs parameter
       in SSL connections.
    """
    return DEFAULT_CA_CERT_PATHS


class HTTPTimeout(object):  # pylint: disable=too-few-public-methods
    """HTTP timeout class that is a context manager (for use by 'with'
    statement).

    Usage:

      ::

        with HTTPTimeout(timeout, http_conn, conn_id):
            ... operations using http_conn ...

    If the timeout expires, the socket of the HTTP connection is shut down.
    Once the http operations return as a result of that or for other reasons,
    the exit handler of this class raises a :exc:`~pywbem.TimeoutError`
    exception in the thread that executed the ``with`` statement.
    """

    def __init__(self, timeout, http_conn, conn_id=None):
        """Initialize the HTTPTimeout object.

        :Parameters:

          timeout (:term:`number`):
            Timeout in seconds, `None` means no timeout.

          http_conn (`httplib.HTTPBaseConnection` or subclass):
            The connection that is to be stopped when the timeout expires.

          conn_id (:term:`connection id`): Connection ID of the WBEM connection
            in whose context the error happened. `None` if the error did not
            happen in context of any connection, or if the connection context
            was not known.
        """

        self._timeout = timeout
        self._http_conn = http_conn
        self._conn_id = conn_id

        # time in seconds after which a retry of socket shutdown is scheduled
        # if the socket is not yet connected when timeout expires
        self._retrytime = 5

        self._timer = None      # the timer object
        self._ts1 = None        # timestamp when timer was started
        self._shutdown = None   # flag indicating timer handler shutdown socket
        return

    def __enter__(self):
        if self._timeout is not None:
            self._timer = threading.Timer(self._timeout,
                                          HTTPTimeout.timer_expired, [self])
            self._timer.start()
            self._ts1 = datetime.now()
        self._shutdown = False
        return

    def __exit__(self, exc_type, exc_value, traceback):
        if self._timeout is not None:
            self._timer.cancel()
            if self._shutdown:
                # If the timer handler has shut down the socket, we
                # want to make that known, and override any other
                # exceptions that may be pending.
                ts2 = datetime.now()
                duration = ts2 - self._ts1
                duration_sec = (float(duration.microseconds) / 1000000) + \
                    duration.seconds + (duration.days * 24 * 3600)
                raise TimeoutError(
                    _format("The client timed out and closed the socket "
                            "after {0:.0f} s.", duration_sec),
                    conn_id=self._conn_id)
        return False  # re-raise any other exceptions

    def timer_expired(self):
        """
        This method is invoked in context of the timer thread, so we cannot
        directly throw exceptions (we can, but they would be in the wrong
        thread), so instead we shut down the socket of the connection.
        When the timeout happens in early phases of the connection setup,
        there is no socket object on the HTTP connection yet, in that case
        we retry after the retry duration, indefinitely.
        So we do not guarantee in all cases that the overall operation times
        out after the specified timeout.
        """
        if self._http_conn.sock is not None:
            self._shutdown = True
            self._http_conn.sock.shutdown(socket.SHUT_RDWR)
        else:
            # Retry after the retry duration
            self._timer.cancel()
            self._timer = threading.Timer(self._retrytime,
                                          HTTPTimeout.timer_expired, [self])
            self._timer.start()


def parse_url(url, allow_defaults=True):
    """Return a tuple (`host`, `port`, `ssl`) from the URL specified in the
    `url` parameter.

    The returned `ssl` item is a boolean indicating the use of SSL, and is
    recognized from the URL scheme (http vs. https). If none of these schemes
    is specified in the URL, the returned value defaults to False
    (non-SSL/http).

    The returned `port` item is the port number, as an integer. If there is
    no port number specified in the URL, the returned value defaults to 5988
    for non-SSL/http, and to 5989 for SSL/https.

    The returned `host` item is the host portion of the URL, as a string.
    The host portion may be specified in the URL as a short or long host name,
    dotted IPv4 address, or bracketed IPv6 address with or without zone index
    (aka scope ID). An IPv6 address is converted from the RFC6874 URI syntax
    to the RFC4007 text representation syntax before being returned, by
    removing the brackets and converting the zone index (if present) from
    "-eth0" to "%eth0".

    Examples for valid URLs can be found in the test program
    `tests/unittest/pywbem/test_cim_http.py`.

    Parameters:

      url:

      allow_defaults - If `True` (default) allow defaults for scheme and
        port. If `False`, raise exception for invalid or missing scheme or
        port.

    Returns:

      tuple of (`host`, `port`, `ssl`)

    Raises:

      ValueError: Exception raised if allow_defaults = False and either
        scheme or port are invalid or missing
    """

    default_ssl = False             # default SSL use (for no or unknown scheme)

    # Look for scheme.
    matches = re.match(r"^(https?)://(.*)$", url, re.I)
    _scheme = None
    if matches:
        _scheme = matches.group(1).lower()
        hostport = matches.group(2)
        ssl = (_scheme == 'https')
    else:
        if not allow_defaults:
            raise ValueError(
                _format("URL {0!A} invalid scheme component", url))
        # The URL specified no scheme (or a scheme other than the expected
        # schemes, but we don't check)
        ssl = default_ssl
        hostport = url

    # Remove trailing path segments, if any.
    # Having URL components other than just slashes (e.g. '#' or '?') is not
    # allowed (but we don't check).
    result = hostport.find("/")
    if result >= 0:
        hostport = hostport[0:result]

    # Look for port.
    # This regexp also works for (colon-separated) IPv6 addresses, because they
    # must be bracketed in a URL.
    matches = re.search(r":([0-9]+)$", hostport)
    if matches:
        host = hostport[0:matches.start(0)]
        port = int(matches.group(1))
    else:
        if not allow_defaults:
            raise ValueError(
                _format("URL {0!A} invalid host/port component", url))
        host = hostport
        port = DEFAULT_PORT_HTTPS if ssl else DEFAULT_PORT_HTTP

    # Reformat IPv6 addresses from RFC6874 URI syntax to RFC4007 text
    # representation syntax:
    #   - Remove the brackets.
    #   - Convert the zone index (aka scope ID) from "-eth0" to "%eth0".
    # Note on the regexp below: The first group needs the '?' after '.+' to
    # become non-greedy; in greedy mode, the optional second group would never
    # be matched.
    matches = re.match(r"^\[(.+?)(?:-(.+))?\]$", host)
    if matches:
        # It is an IPv6 address
        host = matches.group(1)
        if matches.group(2) is not None:
            # The zone index is present
            host += "%" + matches.group(2)

    return host, port, ssl


def get_default_ca_certs():
    """
    Try to find out system path with ca certificates. This path is cached and
    returned. If no path is found out, None is returned.
    """
    # pylint: disable=protected-access
    if not hasattr(get_default_ca_certs, '_path'):
        for path in get_default_ca_cert_paths():
            if os.path.exists(path):
                get_default_ca_certs._path = path
                break
        else:
            get_default_ca_certs._path = None
    return get_default_ca_certs._path


# pylint: disable=too-many-branches,too-many-statements,too-many-arguments
def wbem_request(url, data, creds, cimxml_headers=None, debug=False, x509=None,
                 verify_callback=None, ca_certs=None,
                 no_verification=False, timeout=None, recorders=None,
                 conn_id=None):
    # pylint: disable=too-many-arguments,unused-argument
    # pylint: disable=too-many-locals
    """
    Send an HTTP or HTTPS request to a WBEM server and return the response.

    This function uses Python's built-in `httplib` module.

    Parameters:

      url (:term:`string`):
        URL of the WBEM server (e.g. ``"https://10.11.12.13:6988"``).
        For details, see the `url` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      data (:term:`string`):
        The CIM-XML formatted data to be sent as a request to the WBEM server.

      creds:
        Credentials for authenticating with the WBEM server.
        For details, see the `creds` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      cimxml_headers (:term:`py:iterable` of tuple(string,string)):
        Where each tuple contains: header name, header value.

        CIM-XML extension header fields for the request. The header value
        is a string (unicode or binary) that is not encoded, and the two-step
        encoding required by DSP0200 is performed inside of this function.

        A value of `None` is treated like an empty iterable.

      debug (:class:`py:bool`):
        Boolean indicating whether to create debug information (ignored).

      x509:
        Used for HTTPS with certificates.
        For details, see the `x509` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      verify_callback:
        Used for HTTPS with certificates but only for python 2. Ignored with
        python 3 since the  python 3 ssl implementation does not implement
        any callback mechanism so setting this variable gains the
        user nothing.

        For details, see the `verify_callback` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      ca_certs:
        Used for HTTPS with certificates.
        For details, see the `ca_certs` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      no_verification:
        Used for HTTPS with certificates.
        For details, see the `no_verification` parameter of
        :meth:`~pywbem.WBEMConnection.__init__`.

      timeout (:term:`number`):
        Timeout in seconds, for requests sent to the server. If the server did
        not respond within the timeout duration, the socket for the connection
        will be closed, causing a :exc:`~pywbem.TimeoutError` to be raised.
        A value of `None` means there is no timeout.
        A value of `0` means the timeout is very short, and does not really
        make any sense.
        Note that not all situations can be handled within this timeout, so
        for some issues, this method may take longer to raise an exception.

      recorders (List of :class:`~pywbem.BaseOperationRecorder`):
        List of enabled operation recorders, into which the HTTP request and
        HTTP response will be staged as attributes.

      conn_id (:term:`string`)
        string that uniquely defines a connection.  Used as part of any
        logs created.

    Returns:

        Tuple containing:

            The CIM-XML formatted response data from the WBEM server, as a
            :term:`byte string` object.

            The server response time in seconds as floating point number if
            this data was received from the server. If no data returned
            from server `None` is returned.

    Raises:

        :exc:`~pywbem.AuthError`
        :exc:`~pywbem.ConnectionError`
        :exc:`~pywbem.TimeoutError`
        :exc:`~pywbem.HTTPError`
    """

    class HTTPBaseConnection:  # pylint: disable=no-init
        """ Common base for specific connection classes. Implements
            the send method
        """
        # pylint: disable=too-few-public-methods
        def send(self, strng):
            """
            A copy of `httplib.HTTPConnection.send()`, with these fixes:

            * We fix the problem that the connection gets closed upon error
              32 (EPIPE), by not doing that (If the connection gets closed,
              getresponse() fails). This problem was reported as Python issue
              #5542, and the same fix we do here was integrated into Python
              2.7 and 3.1 or 3.2, but not into Python 2.6 (so we still need
              our fix here).

            * Ensure that the data are bytes, not unicode.
            """
            # NOTE: The attributes come from the httplib mixins in the
            # subclasses so the disable=no-member hides worthless warnings.
            if self.sock is None:  # pylint: disable=no-member
                if self.auto_open:  # pylint: disable=no-member
                    self.connect()  # pylint: disable=no-member
                else:
                    # We raise the same httplib exception the original function
                    # raises. Our caller will handle it.
                    raise httplib.NotConnected()
            if self.debuglevel > 0:  # pylint: disable=no-member
                print(_format("send: {0!A}", strng))
            blocksize = 8192
            if hasattr(strng, 'read') and not isinstance(strng, list):
                if self.debuglevel > 0:  # pylint: disable=no-member
                    print("sendIng a read()able")
                data = strng.read(blocksize)
                while data:
                    # pylint: disable=no-member
                    self.sock.sendall(_ensure_bytes(data))
                    data = strng.read(blocksize)
            else:
                # For unknown reasons, the pylint disable must be on same line:
                self.sock.sendall(_ensure_bytes(strng))  # noqa: E501 pylint: disable=no-member

    class HTTPConnection(HTTPBaseConnection, httplib.HTTPConnection):
        """ Execute client connection without ssl using httplib. """
        def __init__(self, host, port=None, timeout=None):
            # Note: We do not use strict=True in the following call, because it
            # is not clear what side effects that would have, and if no status
            # line comes back we'll certainly find out about that.
            httplib.HTTPConnection.__init__(self, host=host, port=port,
                                            timeout=timeout)

    class HTTPSConnection(HTTPBaseConnection, httplib.HTTPSConnection):
        """ Execute client connection with ssl using httplib."""
        # pylint: disable=R0913,too-many-arguments
        def __init__(self, host, port=None, key_file=None, cert_file=None,
                     ca_certs=None, verify_callback=None, timeout=None):
            # Note: We do not use strict=True in the following call, because it
            # is not clear what side effects that would have, and if no status
            # line comes back we'll certainly find out about that.
            httplib.HTTPSConnection.__init__(self, host=host, port=port,
                                             key_file=key_file,
                                             cert_file=cert_file,
                                             timeout=timeout)
            self.ca_certs = ca_certs
            self.verify_callback = verify_callback
            # issue 297: Verify_callback is  not used in py 3
            if verify_callback is not None and six.PY3:
                warnings.warn("The 'verify_callback' parameter was specified "
                              "on WBEMConnection() but is not used on "
                              "Python 3", UserWarning)

        def connect(self):
            # pylint: disable=too-many-branches
            """Connect to a host on a given (SSL) port."""

            # Connect for M2Crypto ssl package
            if _HAVE_M2CRYPTO:
                # Calling httplib.HTTPSConnection.connect(self) does not work
                # because of its ssl.wrap_socket() call. So we copy the code of
                # that connect() method modulo the ssl.wrap_socket() call.

                # Another change is that we do not pass the timeout value
                # on to the socket call, because that does not work with
                # M2Crypto.

                if sys.version_info[0:2] >= (2, 7):
                    # the source_address parameter was added in Python 2.7
                    self.sock = socket.create_connection(
                        (self.host, self.port), None, self.source_address)
                else:
                    self.sock = socket.create_connection(
                        (self.host, self.port), None)

                # Removed code for tunneling support.

                # End of code from httplib.HTTPSConnection.connect(self).

                ctx = SSL.Context('sslv23')

                if self.cert_file:
                    ctx.load_cert(self.cert_file, keyfile=self.key_file)
                if self.ca_certs:
                    ctx.set_verify(
                        SSL.verify_peer | SSL.verify_fail_if_no_peer_cert,
                        depth=9, callback=verify_callback)
                    if os.path.isdir(self.ca_certs):
                        ctx.load_verify_locations(capath=self.ca_certs)
                    else:
                        ctx.load_verify_locations(cafile=self.ca_certs)
                try:
                    self.sock = SSL.Connection(ctx, self.sock)

                    # Below is a body of SSL.Connection.connect() method
                    # except for the first line (socket connection).

                    # Removed code for tunneling support.

                    # Setting the timeout on the input socket does not work
                    # with M2Crypto, with such a timeout set it calls a
                    # different low level function (nbio instead of bio)
                    # that does not work. The symptom is that reading the
                    # response returns None.
                    # Therefore, we set the timeout at the level of the outer
                    # M2Crypto socket object.
                    # pylint: disable=using-constant-test

                    if self.timeout is not None:
                        self.sock.set_socket_read_timeout(
                            SSL.timeout(self.timeout))
                        self.sock.set_socket_write_timeout(
                            SSL.timeout(self.timeout))

                    self.sock.addr = (self.host, self.port)
                    self.sock.setup_ssl()
                    self.sock.set_connect_state()
                    ret = self.sock.connect_ssl()
                    if self.ca_certs:
                        check = getattr(self.sock, 'postConnectionCheck',
                                        self.sock.clientPostConnectionCheck)
                        if check is not None:
                            if not check(self.sock.get_peer_cert(), self.host):
                                raise ConnectionError(
                                    'SSL error: post connection check failed',
                                    conn_id=conn_id)
                    return ret

                except (SSLError, SSL.SSLError,
                        SSL.Checker.SSLVerificationError) as arg:
                    raise ConnectionError(
                        _format("SSL error {0}: {1}", arg.__class__, arg),
                        conn_id=conn_id)

            # Connect using Python SSL module
            else:
                # Setup the socket context

                # Note: PROTOCOL_SSLv23 allows talking to servers with TLS but
                # not with SSL. For details, see the table in
                # https://docs.python.org/3/library/ssl.html#ssl.wrap_socket
                # Within the defined set of protocol versions, SSLv23 selects
                # the highest protocol version that both client and server
                # support.
                # Issue #893: Consider the use of default_context()
                ctx = SSL.SSLContext(SSL.PROTOCOL_SSLv23)

                if self.cert_file:
                    ctx.load_cert_chain(self.cert_file, keyfile=self.key_file)
                if self.ca_certs:
                    # We need to use CERT_REQUIRED to require that the server
                    # certificate is being validated by the client (against the
                    # certificates in ca_certs).
                    ctx.verify_mode = SSL.CERT_REQUIRED
                    if os.path.isdir(self.ca_certs):
                        ctx.load_verify_locations(capath=self.ca_certs)
                    else:
                        ctx.load_verify_locations(cafile=self.ca_certs)
                    ctx.check_hostname = True
                else:
                    ctx.check_hostname = False
                    ctx.verify_mode = SSL.CERT_NONE

                # setup the socket
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(self.timeout)

                try:
                    self.sock = ctx.wrap_socket(sock,
                                                server_hostname=self.host)
                    return self.sock.connect((self.host, self.port))

                except SSLError as arg:
                    raise ConnectionError(
                        _format("SSL error {0}: {1}", arg.__class__, arg),
                        conn_id=conn_id)
                except CertificateError as arg:
                    raise ConnectionError(
                        _format("SSL certificate error {0}: {1}",
                                arg.__class__, arg),
                        conn_id=conn_id)

    class FileHTTPConnection(HTTPBaseConnection, httplib.HTTPConnection):
        """Execute client connection based on a unix domain socket. """

        def __init__(self, uds_path):
            httplib.HTTPConnection.__init__(self, host='localhost')
            self.uds_path = uds_path

        def connect(self):
            try:
                socket_af = socket.AF_UNIX
            except AttributeError:
                raise ConnectionError(
                    _format("file URLs not supported on {0} platform due to "
                            "missing AF_UNIX support", platform.system()),
                    conn_id=conn_id)
            self.sock = socket.socket(socket_af, socket.SOCK_STREAM)
            self.sock.connect(self.uds_path)

    if not cimxml_headers:
        cimxml_headers = []

    host, port, use_ssl = parse_url(_ensure_unicode(url))

    key_file = None
    cert_file = None

    if use_ssl and x509 is not None:
        cert_file = x509.get('cert_file')
        key_file = x509.get('key_file')

    local_auth_header = None

    # Make sure the data parameter is converted to a UTF-8 encoded byte string.
    # This is important because according to RFC2616, the Content-Length HTTP
    # header must be measured in Bytes (and the Content-Type header will
    # indicate UTF-8).
    data = _ensure_bytes(data)

    data = b'<?xml version="1.0" encoding="utf-8" ?>\n' + data

    # Note that certs get passed even if ca_certs is None and
    # no_verification=False
    if not no_verification and ca_certs is None:
        ca_certs = get_default_ca_certs()
    elif no_verification:
        ca_certs = None

    local = False
    svr_resp_time = None
    if use_ssl:
        client = HTTPSConnection(host=host,
                                 port=port,
                                 key_file=key_file,
                                 cert_file=cert_file,
                                 ca_certs=ca_certs,
                                 verify_callback=verify_callback,
                                 timeout=timeout)
    else:
        if url.startswith('http'):
            client = HTTPConnection(host=host, port=port, timeout=timeout)
        else:
            if url.startswith('file:'):
                url_ = url[5:]
            else:
                url_ = url
            try:
                status = os.stat(url_)
                if S_ISSOCK(status.st_mode):
                    client = FileHTTPConnection(url_)
                    local = True
                else:
                    raise ConnectionError(
                        _format("File URL is not a socket: {0!A}", url),
                        conn_id=conn_id)
            except OSError as exc:
                raise ConnectionError(
                    _format("Error with file URL {0!A}: {1}", url, exc),
                    conn_id=conn_id)

    locallogin = None
    if host in ('localhost', 'localhost6', '127.0.0.1', '::1'):
        local = True
    if local:
        try:
            locallogin = getpass.getuser()
        except (KeyError, ImportError):
            locallogin = None

    method = 'POST'
    target = '/cimom'

    if recorders:
        for recorder in recorders:
            recorder.stage_http_request(conn_id, 11, url, target, method,
                                        dict(cimxml_headers), data)

            # We want clean response data when an exception is raised before
            # the HTTP response comes in:
            recorder.stage_http_response1(conn_id, None, None, None, None)
            recorder.stage_http_response2(None)

    with HTTPTimeout(timeout, client, conn_id):

        try_limit = 5  # Number of tries with authentication challenges.

        for num_tries in range(0, try_limit):  # pylint: disable=unused-variable

            client.putrequest(method, target)

            standard_headers = [
                ('Content-type', 'application/xml; charset="utf-8"'),
                ('Content-length', str(len(data))),
            ]
            if local_auth_header:
                standard_headers.append(local_auth_header)
            elif creds is not None:
                auth = _format("{0}:{1}", creds[0], creds[1])
                auth64 = _ensure_unicode(base64.b64encode(
                    _ensure_bytes(auth))).replace('\n', '')
                standard_headers.append(
                    ('Authorization', 'Basic {0}'.format(auth64)))
            elif locallogin is not None:
                standard_headers.append(
                    ('PegasusAuthorization', 'Local "{0}"'.format(locallogin)))

            # Note: RFC2616 does not permit the use percent-escaping for
            # the standard header fields. It allows octets (except CTL) for
            # the field values. DSP0200 requires the use of UTF-8 encoding
            # followed by percent-encoding for its extension headers.
            # Therefore, we don't encode the standard headers but do encode
            # the CIM-XML extension headers.
            for n, v in standard_headers:
                client.putheader(n, v)
            for n, v in cimxml_headers:
                v = _ensure_unicode(v)
                v = urllib.parse.quote(v)
                client.putheader(n, v)

            try:

                # See RFC 2616 section 8.2.2
                # An http server is allowed to send back an error (presumably
                # a 401), and close the connection without reading the entire
                # request.  A server may do this to protect itself from a DoS
                # attack.
                #
                # If the server closes the connection during our send(), we
                # will either get a socket exception 104 (ECONNRESET:
                # connection reset), or a socket exception 32 (EPIPE: broken
                # pipe).  In either case, thanks to our fixed HTTPConnection
                # classes, we'll still be able to retrieve the response so
                # that we can read and respond to the authentication challenge.
                try:
                    # endheaders() is the first method in this sequence that
                    # actually sends something to the server (using send()).
                    client.endheaders()
                    client.send(data)
                except SocketErrors as exc:
                    if exc.args[0] == errno.ECONNRESET:
                        warnings.warn("Ignoring socket error ECONNRESET "
                                      "(connection reset), continuing with "
                                      "reading the response.",
                                      ToleratedServerIssueWarning,
                                      stacklevel=2)
                        # continue with reading response
                    elif exc.args[0] == errno.EPIPE:
                        warnings.warn("Ignoring socket error EPIPE "
                                      "(broken pipe), continuing with "
                                      "reading the response.",
                                      ToleratedServerIssueWarning,
                                      stacklevel=2)
                        # continue with reading response
                    else:
                        raise

                response = client.getresponse()

                # Attempt to get the optional response time header sent from
                # the server
                svr_resp_time = response.getheader(
                    'WBEMServerResponseTime', None)
                if svr_resp_time:
                    try:
                        # convert to float and map from microsec to sec.
                        svr_resp_time = float(svr_resp_time) / 1000000
                    except ValueError:
                        pass

                if recorders:
                    for recorder in recorders:
                        recorder.stage_http_response1(
                            conn_id,
                            response.version,
                            response.status,
                            response.reason,
                            dict(response.getheaders()))

                if response.status != 200:
                    if response.status == 401:
                        if not local:
                            raise AuthError(response.reason, conn_id=conn_id)
                        auth_chal = response.getheader('WWW-Authenticate', '')
                        if 'openwbem' in response.getheader('Server', ''):
                            if 'OWLocal' not in auth_chal:
                                try:
                                    uid = os.getuid()
                                except AttributeError:
                                    raise AuthError(
                                        _format(
                                            "OWLocal authorization for "
                                            "OpenWbem server not supported on "
                                            "{0} platform due to missing "
                                            "os.getuid()", platform.system()),
                                        conn_id=conn_id)
                                local_auth_header = (
                                    'Authorization',
                                    'OWLocal uid="{0}"'.format(uid))
                                continue  # with next retry
                            else:
                                try:
                                    nonce_idx = auth_chal.index('nonce=')
                                    nonce_begin = auth_chal.index('"',
                                                                  nonce_idx)
                                    nonce_end = auth_chal.index('"',
                                                                nonce_begin + 1)
                                    nonce = auth_chal[nonce_begin + 1:nonce_end]
                                    cookie_idx = auth_chal.index('cookiefile=')
                                    cookie_begin = auth_chal.index('"',
                                                                   cookie_idx)
                                    cookie_end = auth_chal.index(
                                        '"', cookie_begin + 1)
                                    cookie_file = auth_chal[
                                        cookie_begin + 1: cookie_end]
                                    file_hndl = open(cookie_file, 'r')
                                    cookie = file_hndl.read().strip()
                                    file_hndl.close()
                                    local_auth_header = (
                                        'Authorization',
                                        'OWLocal nonce="{0}", cookie="{1}"'.
                                        format(nonce, cookie))
                                    continue  # with next retry
                                # pylint: disable=broad-except
                                except Exception as exc:
                                    local_auth_header = None
                                    warnings.warn(
                                        _format(
                                            "Exception in OpenWBEM auth "
                                            "challenge processing: {0} "
                                            "(retrying - this was attempt "
                                            "{1} of {2})",
                                            exc, num_tries + 1, try_limit),
                                        ToleratedServerIssueWarning,
                                        stacklevel=2)
                                    continue  # with next retry
                        elif 'Local' in auth_chal:
                            try:
                                beg = auth_chal.index('"') + 1
                                end = auth_chal.rindex('"')
                                if end > beg:
                                    _file = auth_chal[beg:end]
                                    file_hndl = open(_file, 'r')
                                    cookie = file_hndl.read().strip()
                                    file_hndl.close()
                                    local_auth_header = (
                                        'PegasusAuthorization',
                                        'Local "{0}:{1}:{2}"'.
                                        format(locallogin, _file, cookie))
                                    continue  # with next retry
                            except ValueError:
                                pass
                        raise AuthError(response.reason, conn_id=conn_id)

                    cimerror_hdr = response.getheader('CIMError', None)
                    if cimerror_hdr is not None:
                        cimdetails = {}
                        pgdetails_hdr = response.getheader('PGErrorDetail',
                                                           None)
                        if pgdetails_hdr is not None:
                            # pylint: disable=too-many-function-args
                            cimdetails['PGErrorDetail'] = \
                                urllib.parse.unquote(pgdetails_hdr)
                        raise HTTPError(response.status, response.reason,
                                        cimerror_hdr, cimdetails,
                                        conn_id=conn_id)

                    raise HTTPError(response.status, response.reason,
                                    conn_id=conn_id)

                body = response.read()

                if recorders:
                    for recorder in recorders:
                        recorder.stage_http_response2(body)

            except httplib.BadStatusLine as exc:
                # Background: BadStatusLine is documented to be raised only
                # when strict=True is used (that is not the case here).
                # However, httplib currently raises BadStatusLine also
                # independent of strict when a keep-alive connection times out
                # (e.g. because the server went down).
                # See https://bugs.python.org/issue8450.
                # In Python 3.5, this case is raised as a new
                # httplib.RemoteDisconnected exception, which is also still
                # a BadStatusLine exception with an empty line.
                if exc.line is None or exc.line.strip().strip("'") in \
                        ('', 'None'):
                    # TODO 4/2018 AM Enable retry logic. For unknown reasons,
                    #   retrying causes testclient test case SocketError104 to
                    #   fail. Also, retrying needs to be tested with a real
                    #   WBEM server.
                    if False:  # pylint: disable=using-constant-test
                        warnings.warn(
                            _format("The server closed the connection without "
                                    "returning any response (retrying - this "
                                    "was attempt {0} of {1} at {2})",
                                    num_tries + 1, try_limit, datetime.now()),
                            ToleratedServerIssueWarning,
                            stacklevel=2)
                        continue  # with next retry
                    else:
                        raise ConnectionError(
                            "The server closed the connection without "
                            "returning any response",
                            conn_id=conn_id)
                else:
                    raise ConnectionError(
                        _format("The server returned a bad HTTP status line: "
                                "{0!A}", exc.line),
                        conn_id=conn_id)
            except httplib.IncompleteRead as exc:
                raise ConnectionError(
                    _format("HTTP incomplete read: {0}", exc),
                    conn_id=conn_id)
            except httplib.NotConnected as exc:
                raise ConnectionError(
                    _format("HTTP not connected: {0}", exc),
                    conn_id=conn_id)
            except httplib.HTTPException as exc:
                # Base class for all httplib exceptions
                raise ConnectionError(
                    _format("HTTP error: {0}", exc),
                    conn_id=conn_id)
            except SocketErrors as exc:
                raise ConnectionError(
                    _format("Socket error: {0}", exc),
                    conn_id=conn_id)

            # Operation was successful
            break

    return body, svr_resp_time


def get_cimobject_header(obj):
    """
    Return the value for the CIM-XML extension header field 'CIMObject', using
    the given object.

    This function implements the rules defined in DSP0200 section 6.3.7
    "CIMObject". The format of the CIMObject value is similar but not identical
    to a local WBEM URI (one without namespace type and authority), as defined
    in DSP0207.

    One difference is that DSP0207 requires a leading slash for a local WBEM
    URI, e.g. '/root/cimv2:CIM_Class.k=1', while the CIMObject value has no
    leading slash, e.g. 'root/cimv2:CIM_Class.k=1'.

    Another difference is that the CIMObject value for instance paths has
    provisions for an instance path without keys, while WBEM URIs do not have
    that. Pywbem does not support that.
    """

    # Local namespace path
    if isinstance(obj, six.string_types):
        return obj

    # Local class path
    if isinstance(obj, CIMClassName):
        return obj.to_wbem_uri(format='cimobject')

    # Local instance path
    if isinstance(obj, CIMInstanceName):
        return obj.to_wbem_uri(format='cimobject')

    raise TypeError(
        _format("Invalid object type {0} to generate CIMObject header value "
                "from", type(obj)))



"""
Pytest fixtures for pywbem end2end tests.
"""

from __future__ import absolute_import

import os
import warnings
import pytest
import yaml

from pywbem import WBEMServer, Error, ConnectionError, AuthError, \
    ToleratedServerIssueWarning

from .server_definition_file import ServerDefinition, \
    ServerDefinitionFile
from .utils import latest_profile_inst, ServerObjectCache, \
    WBEMConnectionAsserted, server_func_asserted
from .assertions import assert_association_a1, assert_association_a2, \
    assert_association_a3, assert_association_a4, assert_association_a5, \
    assert_association_a6

# Server nickname or server group nickname in WBEM server definition file
TESTSERVER = os.getenv('TESTSERVER', 'default')
SD_LIST = ServerDefinitionFile().list_servers(TESTSERVER)

# Profile definition file (in end2end directory).
PROFILES_YAML_FILE = os.path.join('tests', 'profiles', 'profiles.yml')

# Profile definition list.
# The list items are profile definition items, as described in the profile
# definition file.
with open(PROFILES_YAML_FILE, 'r') as _fp:
    PROFILE_DEFINITION_LIST = yaml.load(_fp)
del _fp  # pylint: disable=undefined-loop-variable

# Profile definition dictionary.
# The dict key is 'org:name', for optimized direct access.
# the dict values are profile definition items, as described in the profile
# definition file.
PROFILE_DEFINITION_DICT = dict()
for _pd in PROFILE_DEFINITION_LIST:
    _key = '{0}:{1}'.format(_pd['registered_org'],
                            _pd['registered_name'])
    PROFILE_DEFINITION_DICT[_key] = _pd
del _pd, _key  # pylint: disable=undefined-loop-variable


def fixtureid_server_definition(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `server_definition()`.

    Parameters:
      * fixture_value (ServerDefinition): The server definition of the
        WBEM server the test runs against.
    """
    sd = fixture_value
    assert isinstance(sd, ServerDefinition)
    return "server_definition={0}".format(sd.nickname)


@pytest.fixture(
    params=SD_LIST,
    scope='module',
    ids=fixtureid_server_definition
)
def server_definition(request):
    """
    Fixture representing the set of WBEM server definitions to use for the
    end2end tests.

    Returns the `ServerDefinition` object of each server to test against.
    """
    return request.param


@pytest.fixture(
    scope='module'
)
def wbem_connection(request, server_definition):
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Fixture representing the set of WBEMConnection objects to use for the
    end2end tests.  This fixture uses the  server_definition
    implementation_namespace item from the server_definition as the
    default namespace if that item exists in the server_definition.

    Returns the `WBEMConnection` object of each server to test against.
    """
    sd = server_definition

    # Skip this server if we had a skip reason in an earlier attempt
    skip_msg = getattr(sd, 'skip_msg', None)
    if skip_msg:
        pytest.skip("Remembered skip reason from earlier attempt: {0}".
                    format(skip_msg))

    x509 = dict(cert_file=sd.cert_file, key_file=sd.key_file) \
        if sd.cert_file and sd.key_file else None

    # Define the WBEMConnection default_namespace as the
    # implementation_namespace if that data exists, else define it as
    # None which invokes root/cimv2 as the default namespace.
    # NOTE: Since WBEMConnection assigns root/cimv2 as default, there is
    # no way to really know if a valid implementation namespace was defined.
    conn = WBEMConnectionAsserted(
        sd.url, (sd.user, sd.password),
        x509=x509,
        ca_certs=sd.ca_certs,
        no_verification=sd.no_verification,
        default_namespace=sd.implementation_namespace,
        timeout=10)

    conn.server_definition = sd

    # Check that the server can be reached and authenticated with, by issuing
    # some quick operation. Operation failures are tolerated (e.g. GetQualifier
    # is not supported on SFCB), and the connection and authgentication errors
    # can still be detected because they will be detected before the operation
    # fails.
    try:
        conn.GetQualifier(
            'Association',
            asserted=False)
    except ConnectionError as exc:
        msg = "Server {0} at {1}: Server cannot be reached: {2} - {3}". \
            format(sd.nickname, sd.url, exc.__class__.__name__, exc)
        sd.skip_msg = msg
        warnings.warn(msg, ToleratedServerIssueWarning)
        pytest.skip(msg)
    except AuthError as exc:
        msg = "Server {0} at {1}: Server cannot be authenticated with: " \
            "{2} - {3}". \
            format(sd.nickname, sd.url, exc.__class__.__name__, exc)
        sd.skip_msg = msg
        warnings.warn(msg, ToleratedServerIssueWarning)
        pytest.skip(msg)
    except Error:
        # Any other error may be due to the WBEM server not supporting
        # qualifier operations. In that case, we don't optimize.
        sd.skip_msg = None
    else:
        sd.skip_msg = None

    return conn


def fixtureid_default_namespace(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `default_namespace()`.

    Parameters:
      * fixture_value (string): The default namespace for the test.
    """
    ns = fixture_value
    return "default_namespace={0}".format(ns)


@pytest.fixture(
    params=[None, 'root/cimv2', 'interop', 'root/interop'],
    scope='module',
    ids=fixtureid_default_namespace
)
def default_namespace(request):
    """
    Fixture representing the set of default namespaces to open connections
    with.

    Returns the default namespace as a string.
    """
    return request.param


def fixtureid_profile_definition(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `profile_definition()`.

    Parameters:
      * fixture_value (dict): The profile definition dictionary of the profile.
    """
    pd = fixture_value
    assert isinstance(pd, dict)
    assert 'registered_name' in pd
    assert 'registered_org' in pd
    return "profile_definition={0}:{1}". \
        format(pd['registered_org'], pd['registered_name'])


def _apply_profile_definition_defaults(pd):
    """Set the reference_direction parameter"""
    if 'reference_direction' not in pd:
        pd['reference_direction'] = None
    if pd['reference_direction'] is None:
        org = pd['registered_org']
        if org == 'SNIA':
            pd['reference_direction'] = 'snia'
        elif org == 'DMTF':
            pd['reference_direction'] = 'dmtf'
        else:
            raise ValueError(
                "Profile definition error: The reference_direction item of a "
                "profile definition can only be defaulted when the registered "
                "organisation is DMTF or SNIA, but it is {0}".format(org))
    if 'scoping_class' not in pd:
        pd['scoping_class'] = None
    if 'scoping_path' not in pd:
        pd['scoping_path'] = None
    return pd


def single_profile_definition(org, name):
    """
    Return a single profile definition dictionary for the specified org and
    name. If a profile definition does not exist, returns None.

    Optional items in the profile definition dictionary are generated with
    their default values.
    """
    assert org is not None
    assert name is not None
    pd_key = "{0}:{1}".format(org, name)
    try:
        pd = PROFILE_DEFINITION_DICT[pd_key]
    except KeyError:
        return None
    pd = _apply_profile_definition_defaults(pd.copy())
    return pd


@pytest.fixture(
    params=PROFILE_DEFINITION_LIST,
    scope='module',
    ids=fixtureid_profile_definition
)
def profile_definition(request):
    """
    Fixture representing the set of profile definitions to use for the
    end2end tests.

    Returns the profile definition dictionary of each profile to test against.
    For dict items, see the description in profiles.yaml.

    Optional items in the dictionary are generated with their default values.
    """
    pd = _apply_profile_definition_defaults(request.param.copy())
    return pd


@pytest.fixture(
    params=[
        # the function for a1 must be first, because its results are stored
        # in the object cache for use by the subsequent functions.
        assert_association_a1,
        assert_association_a2,
        assert_association_a3,
        assert_association_a4,
        assert_association_a5,
        assert_association_a6,
    ],
    scope='module',
)
def assert_association_func(request):
    """
    Fixture representing a function for asserting association traversal.

    Returns the function.

    Interface of the function:

        Parameters:
          conn, profile_id, source_path, source_role, assoc_class, far_role,
          far_class

        Returns:
            tuple of:
            - list of associated (far end) instances
            - list of association instances
    """
    return request.param


class ProfileTest(object):
    """
    Base class for end2end tests on a specific profile.
    """

    object_cache = ServerObjectCache()

    def init_profile(self, conn, profile_org=None, profile_name=None):
        """
        Initialize attributes for the profile.

        The following instance attributes are set:
          * conn: conn argument (WBEMConnection).
          * profile_org: profile_org argument (registered org string).
          * profile_name: profile_name argument (registered name string).
          * server: WBEMServer object created from the connection.
          * profile_definition: Profile definition dictionary for the profile.
          * profile_id: org:name string to identify the profile
          * profile_inst: CIMInstance for the CIM_RegisteredProfile object for
            the profile. If it does not exist, the testcase is skipped.
        """
        # pylint: disable=attribute-defined-outside-init
        self.conn = conn
        self.profile_org = profile_org
        self.profile_name = profile_name

        self.server = WBEMServer(conn)
        self.profile_definition = single_profile_definition(
            profile_org, profile_name)
        assert self.profile_definition is not None

        self.profile_id = "{0}:{1}".format(profile_org, profile_name)
        profile_insts_id = self.profile_id + ':profile_insts'

        try:
            profile_insts = self.object_cache.get_list(
                conn.url, profile_insts_id)
        except KeyError:
            profile_insts = server_func_asserted(
                self.server, 'get_selected_profiles',
                profile_org, profile_name)
            self.object_cache.add_list(
                conn.url, profile_insts_id, profile_insts)

        if not profile_insts:
            pytest.skip("{0} {1} profile is not advertised on server "
                        "{2!r}".
                        format(profile_org, profile_name, self.conn.url))

        self.profile_inst = latest_profile_inst(profile_insts)
        # pylint: enable=attribute-defined-outside-init

    def init_central_instances(self, conn):
        """
        Initialize attributes for the profile by calling init_profile()
        and get the central instances.

        The following instance attributes are set, in addition to
        init_profile():
          * central_inst_paths: central instances as list of CIMInstanceName.
        """

        self.init_profile(conn)

        central_paths_id = self.profile_id + ':central_paths'

        try:
            central_paths = self.object_cache.get_list(
                self.conn.url, central_paths_id)
        except KeyError:
            central_paths = server_func_asserted(
                self.server, 'get_central_instances',
                self.profile_inst.path,
                central_class=self.profile_definition['central_class'],
                scoping_class=self.profile_definition['scoping_class'],
                scoping_path=self.profile_definition['scoping_path'],
                reference_direction=self.profile_definition[
                    'reference_direction'])
            self.object_cache.add_list(
                self.conn.url, central_paths_id, central_paths)

        # pylint: disable=attribute-defined-outside-init
        self.central_inst_paths = central_paths



# -*- coding: utf-8 -*-
#
# Configuration file for Sphinx builds, created by
# sphinx-quickstart on Wed Mar  2 11:33:06 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import re

from pbr.version import VersionInfo

# Imports used by code for autoautosummary
import inspect
from sphinx.ext.autosummary import Autosummary
from sphinx.ext.autosummary import get_documenter
from docutils.parsers.rst import directives
from sphinx.util.inspect import safe_getattr
from sphinx.util import logging

from pywbem.cim_constants import DEFAULT_NAMESPACE

# RST variable substitutions
rst_prolog = """

.. |DEFAULT_NAMESPACE| replace:: ``"%s"``

""" % DEFAULT_NAMESPACE

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.7'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',   # disabed, raises anexception
    'sphinx.ext.ifconfig',
    'sphinx_git',            # requires 'sphinx-git' Python package
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    master_doc = 'index'
else:
    master_doc = 'docs/index'

# General information about the project.
project = u'pywbem'
#copyright = u''
author = u'pywbem team'

# The short description of the package.
_short_description = u'Pywbem - a WBEM client written in pure Python'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The short X.Y version.
# Note: We use the full version in both cases (e.g. 'M.N.U' or 'M.N.U.dev0').
version = VersionInfo('pywbem').release_string()

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["README.rst", "listener_design.rst",
                    "tests", ".tox", ".git", "attic", "dist", "irecv",
                    "tools", "packaging", "build_doc", "pywbem.egg-info",
                    ".eggs"]

# The reST default role (used for this markup: `text`) to use for all
# documents. None means it is rendered in italic, without a link.
default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for Napoleon extension ---------------------------------------

# Enable support for Google style docstrings. Defaults to True.
napoleon_google_docstring = True

# Enable support for NumPy style docstrings. Defaults to True.
napoleon_numpy_docstring = False

# Include private members (like _membername). False to fall back to Sphinx’s
# default behavior. Defaults to False.
napoleon_include_private_with_doc = False

# Include special members (like __membername__). False to fall back to Sphinx’s
# default behavior. Defaults to True.
napoleon_include_special_with_doc = True

# Use the .. admonition:: directive for the Example and Examples sections,
# instead of the .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_examples = False

# Use the .. admonition:: directive for Notes sections, instead of the
# .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_notes = False

# Use the .. admonition:: directive for References sections, instead of the
# .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_references = False

# Use the :ivar: role for instance variables, instead of the .. attribute::
# directive. Defaults to False.
napoleon_use_ivar = True

# Use a :param: role for each function parameter, instead of a single
# :parameters: role for all the parameters. Defaults to True.
napoleon_use_param = True

# Use the :rtype: role for the return type, instead of inlining it with the
# description. Defaults to True.
napoleon_use_rtype = True


# -- Options for viewcode extension ---------------------------------------

# Follow alias objects that are imported from another module such as functions,
# classes and attributes. As side effects, this option ... ???
# If false, ... ???.
# The default is True.
viewcode_import = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
# See https://www.sphinx-doc.org/en/stable/theming.html for built-in themes.
html_theme = "classic"

# Theme options are theme-specific and customize the look and feel of a theme
# further.
# See https://www.sphinx-doc.org/en/stable/theming.html for the options
# available for built-in themes.
html_theme_options = {
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If not defined, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = 'ld'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = ['_extra']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'pywbem_doc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'pywbem.tex', _short_description, author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pywbem', _short_description, [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'pywbem', _short_description,
     author, 'pywbem', _short_description,
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -- Options for autodoc extension ----------------------------------------
# For documentation, see
# https://www.sphinx-doc.org/en/stable/ext/autodoc.html

# Note on the :special-members: option:
# In Sphinx releases azt least up to 1.6.5, this option does not behave
# as documented. Its behavior is that it only has an effect on the presence
# of the __init__ member in the documentation, which is shown when the
# option is specified without arguments or with an an argument list that
# includes __init__. Other special members that exist in the code are
# always shown (regardless of whether the option is omitted, specified without
# arguments, or specified with an argument list that may or may not
# include the special member).

# Selects what content will be inserted into a class description.
# The possible values are:
#   "class" - Only the class’ docstring is inserted. This is the default.
#   "both"  - Both the class’ and the __init__ method’s docstring are
#             concatenated and inserted.
#   "init"  - Only the __init__ method’s docstring is inserted.
# In all cases, the __init__ method is still independently rendered as a
# special method when the :special-members: option of the autoclass
# directive includes __init__ or is specified with no arguments.
# Based upon the behavior of the :special-members: option described above,
# the recommendation is to not specify the :special-members: option
# when this config value is set to "both" or "init".
autoclass_content = "both"

# Selects if automatically documented members are sorted alphabetically
# (value 'alphabetical'), by member type (value 'groupwise') or by source
# order (value 'bysource'). The default is alphabetical.
autodoc_member_order = "bysource"

# This value is a list of autodoc directive options (flags) that should be
# automatically applied to all autodoc directives. The supported options
# are:
#   'members', 'undoc-members', 'private-members', 'special-members',
#   'inherited-members' and 'show-inheritance'.
# If you set one of these options in this config value, they behave as if
# they had been specified without arguments on each applicable autodoc
# directive. If needed, an autodoc directive can then unspecify the option
# for the current autodoc directive with a negated form :no-{option}:.
# For example, you would specify an option :no-members: on an autoclass
# directive to unspecify a 'members' option included in this config value.
# Note that the :members: option on automodule is recursive w.r.t. the
# classes or other items in the module, so when you want to have specific
# autoclass directives, make sure that the :nmembers: option is not
# set for automodule.
autodoc_default_flags = []

# Functions imported from C modules cannot be introspected, and therefore the
# signature for such functions cannot be automatically determined. However, it
# is an often-used convention to put the signature into the first line of the
# function’s docstring.
# If this boolean value is set to True (which is the default), autodoc will
# look at the first line of the docstring for functions and methods, and if it
# looks like a signature, use the line as the signature and remove it from the
# docstring content.
autodoc_docstring_signature = True

# This value contains a list of modules to be mocked up. This is useful when
# some external dependencies are not met at build time and break the building
# process.
autodoc_mock_imports = []


# -- Options for intersphinx extension ------------------------------------
# For documentation, see
# https://www.sphinx-doc.org/en/stable/ext/intersphinx.html

# Defines the prefixes for intersphinx links, and the targets they resolve to.
# Example RST source for 'py2' prefix:
#     :func:`py2:platform.dist`
#
# Note: The URLs apparently cannot be the same for two different IDs; otherwise
#       the links for one of them are not being created. A small difference
#       such as adding a trailing backslash is already sufficient to work
#       around the problem.
#
# Note: This mapping does not control how links to datatypes of function
#       parameters are generated.
# TODO: Find out how the targeted Python version for auto-generated links
#       to datatypes of function parameters can be controlled.
#
intersphinx_mapping = {
  'py': ('https://docs.python.org/2/', None), # agnostic to Python version
  'py2': ('https://docs.python.org/2', None), # specific to Python 2
  'py3': ('https://docs.python.org/3', None), # specific to Python 3
}

intersphinx_cache_limit = 5

# -- Options for extlinks extension ---------------------------------------
# For documentation, see
# https://www.sphinx-doc.org/en/stable/ext/extlinks.html
#
# Defines aliases for external links that can be used as role names.
#
# This config value must be a dictionary of external sites, mapping unique
# short alias names to a base URL and a prefix:
# * key: alias-name
# * value: tuple of (base-url, prefix)
#
# Example for the config value:
#
#   extlinks = {
#     'issue': ('https://github.com/sphinx-doc/sphinx/issues/%s', 'Issue ')
#   }
#
# The alias-name can be used as a role in links. In the example, alias name
# 'issue' is used in RST as follows:
#   :issue:`123`.
# This then translates into a link:
#   https://github.com/sphinx-doc/sphinx/issues/123
# where the %s in the base-url was replaced with the value between back quotes.
#
# The prefix plays a role only for the link caption:
# * If the prefix is None, the link caption is the full URL.
# * If the prefix is the empty string, the link caption is the partial URL
#   given in the role content ("123" in this case.)
# * If the prefix is a non-empty string, the link caption is the partial URL,
#   prepended by the prefix. In the above example, the link caption would be
#   "Issue 123".
#
# You can also use the usual "explicit title" syntax supported by other roles
# that generate links to set the caption. In this case, the prefix is not
# relevant.
# For example, this RST:
#   :issue:`this issue <123>`
# results in the link caption "this issue".

extlinks = {
  'nbview': ('https://nbviewer.jupyter.org/github/pywbem/pywbem/blob/master/docs/notebooks/%s', ''),
  'nbdown': ('https://github.com/pywbem/pywbem/raw/master/docs/notebooks/%s', '')
}

# -- Support for autoautosummary ----------------------------------------------
#
# Idea taken from https://stackoverflow.com/a/30783465/1424462
#

class AutoAutoSummary(Autosummary):
    """
    Sphinx extension that automatically generates a table of public methods or
    attributes of a class, using the AutoSummary extension.
    (i.e. each row in the table shows the method or attribute name with a
    link to the full description, and a one-line brief description).

    Usage in RST source::

        .. autoclass:: path.to.class
           :<autoclass-options>:

           .. rubric:: Methods

           .. autoautosummary:: path.to.class
              :methods:

           .. rubric:: Attributes

           .. autoautosummary:: path.to.class
              :attributes:

           .. rubric:: Details

    """

    option_spec = {
        'methods': directives.unchanged,
        'attributes': directives.unchanged
    }
    option_spec.update(Autosummary.option_spec)

    required_arguments = 1  # Fully qualified class name

    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(__name__)  # requires Sphinx 1.6.1
        self._log_prefix = "conf.py/AutoAutoSummary"
        self._excluded_classes = ['BaseException']
        super(AutoAutoSummary, self).__init__(*args, **kwargs)

    def _get_members(self, class_obj, member_type, include_in_public=None):
        """
        Return class members of the specified type.

        class_obj: Class object.

        member_type: Member type ('method' or 'attribute').

        include_in_public: set/list/tuple with member names that should be
          included in public members in addition to the public names (those
          starting without underscore).

        Returns:
          tuple(public_members, all_members): Names of the class members of
            the specified member type (public / all).
        """
        try:
            app = self.state.document.settings.env.app
        except AttributeError:
            app = None
        if not include_in_public:
            include_in_public = []
        all_members = []
        for member_name in dir(class_obj):
            try:
                documenter = get_documenter(
                    app,
                    safe_getattr(class_obj, member_name),
                    class_obj)
            except AttributeError:
                continue
            if documenter.objtype == member_type:
                all_members.append(member_name)
        public_members = [x for x in all_members
                  if x in include_in_public or not x.startswith('_')]
        return public_members, all_members

    def _get_def_class(self, class_obj, member_name):
        """
        Return the class object in MRO order that defines a member.

        class_obj: Class object that exposes (but not necessarily defines) the
          member. I.e. starting point of the search.

        member_name: Name of the member (method or attribute).

        Returns:
          Class object that defines the member.
        """
        member_obj = getattr(class_obj, member_name)
        for def_class_obj in inspect.getmro(class_obj):
            if member_name in def_class_obj.__dict__:
                if def_class_obj.__name__ in self._excluded_classes:
                    return class_obj  # Fall back to input class
                return def_class_obj
        self._logger.warning(
            "%s: Definition class not found for member %s.%s, "
            "defaulting to class %s",
            self._log_prefix, class_obj.__name__, member_name,
            class_obj.__name__)
        return class_obj  # Input class is better than nothing

    def run(self):

        try:
            full_class_name = str(self.arguments[0])
            module_name, class_name = full_class_name.rsplit('.', 1)
            module_obj = __import__(module_name, globals(), locals(),
                                    [class_name])
            class_obj = getattr(module_obj, class_name)
            if 'methods' in self.options:
                _, methods = self._get_members(
                    class_obj, 'method', ['__init__'])
                self.content = []
                for method in methods:
                    if method.startswith('_'):
                        # Skip private methods
                        continue
                    def_class = self._get_def_class(class_obj, method)
                    def_module_name = def_class.__module__
                    if def_module_name.startswith('pywbem'):
                        def_module_name = def_module_name.split('.')[0]
                    content_str = "~%s.%s.%s" % (
                        def_module_name,
                        def_class.__name__,
                        method)
                    self.content.append(content_str)
            elif 'attributes' in self.options:
                _, attributes = self._get_members(class_obj, 'attribute')
                self.content = []
                for attrib in attributes:
                    if attrib.startswith('_'):
                        # Skip private attributes
                        continue
                    def_class = self._get_def_class(class_obj, attrib)
                    def_module_name = def_class.__module__
                    if def_module_name.startswith('pywbem'):
                        def_module_name = def_module_name.split('.')[0]
                    content_str = "~%s.%s.%s" % (
                        def_module_name,
                        def_class.__name__,
                        attrib)
                    self.content.append(content_str)

        except Exception as exc:
            self._logger.error(
                "%s: Internal error: %s: %s",
                self._log_prefix, exc.__class__.__name__, exc)

        finally:
            return super(AutoAutoSummary, self).run()


def setup(app):
    app.add_directive('autoautosummary', AutoAutoSummary)
